"""
TODO:
 This data MUST be delivered via dAuth to all nodes in the network.

# Mainnet & Testnet

EE_GENESIS_EPOCH_DATE="2025-02-03 17:00:00"
EE_EPOCH_INTERVALS=24
EE_EPOCH_INTERVAL_SECONDS=3600

# Devnet
EE_GENESIS_EPOCH_DATE="2025-01-24 00:00:00"
EE_EPOCH_INTERVALS=1
EE_EPOCH_INTERVAL_SECONDS=3600


  
"""
import uuid
import json
import os

import numpy as np

from datetime import datetime, timedelta, timezone
from collections import defaultdict, deque
from copy import deepcopy
from threading import Lock
from time import time


from naeural_core import constants as ct
from naeural_core.utils import Singleton

from naeural_core.main.ver import __VER__ as CORE_VERSION
from ratio1 import version as SDK_VERSION

try:
  from ver import __VER__ as NODE_VERSION
except Exception as e:
  NODE_VERSION = 'CORE'

EPOCH_MANAGER_VERSION = '0.3.1'

EPOCH_MAX_VALUE = ct.EPOCH_MAX_VALUE


DEFAULT_NODE_ALERT_INTERVAL = ct.DEFAULT_EPOCH_INTERVAL_SECONDS


SUPERVISOR_MIN_AVAIL_UINT8 = int(ct.SUPERVISOR_MIN_AVAIL_PRC * EPOCH_MAX_VALUE)

FN_NAME = 'epochs_status.pkl'
FN_SUBFOLDER = 'network_monitor'
FN_FULL = FN_SUBFOLDER + '/' + FN_NAME

EPOCHMON_MUTEX = 'epochmon_mutex'


INITIAL_SYNC_EPOCH = 0  # TODO: add initial sync epoch

try:
  EPOCH_MANAGER_DEBUG = int(os.environ.get(ct.EE_EPOCH_MANAGER_DEBUG, 1))
except Exception as e:
  EPOCH_MANAGER_DEBUG = 1

SYNC_SIGNATURES = 'SIGNATURES'
SYNC_AGREEMENT_CID = 'AGREEMENT_CID'
SYNC_LAST_EPOCH = 'LAST_SYNC_EPOCH'
SYNC_NODES = 'NODES'

SYNC_SAVES_TS = 'SAVES_UTC'
SYNC_SAVES_EP = 'SAVE_EPOCHS'
SYNC_RESTARTS = 'EM_RESTARTS_UTC'
SYNC_RELOADS = 'EM_RELOADS_UTC'
FAULTY_EPOCHS = 'FAULTY_EPOCHS'

_FULL_DATA_TEMPLATE_EXTRA = {
  SYNC_LAST_EPOCH : INITIAL_SYNC_EPOCH,
  SYNC_SAVES_TS : [],
  SYNC_SAVES_EP : [],
  SYNC_RESTARTS : [],
  SYNC_RELOADS : [],
  # TODO: should this be in FULL_DATA_MANDATORY_FIELDS?
  SYNC_SIGNATURES : defaultdict(dict),
  SYNC_AGREEMENT_CID : {},
  # In this list will be stored the epochs where consensus was not reached.
  # For those epochs all nodes with a license associated previously
  # will be considered as fully available.
  FAULTY_EPOCHS : [],
  
  ct.EE_GENESIS_EPOCH_DATE_KEY : None,
  ct.BASE_CT.EE_EPOCH_INTERVALS_KEY : None,
  ct.BASE_CT.EE_EPOCH_INTERVAL_SECONDS_KEY : None,
}

_FULL_DATA_MANDATORY_FIELDS = [
  SYNC_NODES,
  ct.EE_GENESIS_EPOCH_DATE_KEY ,
  ct.BASE_CT.EE_EPOCH_INTERVALS_KEY,
  ct.BASE_CT.EE_EPOCH_INTERVAL_SECONDS_KEY ,
]

_FULL_DATA_INFO_KEYS = [
  SYNC_SAVES_TS,
  SYNC_SAVES_EP,
  SYNC_RESTARTS,
  SYNC_RELOADS,
  # TODO: maybe exclude this from info?
  FAULTY_EPOCHS,
]

SYNC_HISTORY_SIZE = 10

class EPCT:
  NAME = 'name'
  ID = 'id'
  EPOCHS = 'epochs'
  ALERTS = 'alerts'
  LAST_ALERT_TS = 'last_alert_ts'
  CURRENT_EPOCH = 'current_epoch'
  HB_TIMESTAMPS = 'hb_dates'
  HB_COUNT = 'hb_count'
  FIRST_SEEN = 'first_seen'
  LAST_SEEN = 'last_seen'
  LAST_EPOCH = 'last_epoch'
  
  LAST_EPOCH_RESTARTS = 'last_epoch_restarts'

  SIGNATURES = 'signatures'
  

_NODE_TEMPLATE = {
  EPCT.NAME           : None,
  EPCT.EPOCHS         : defaultdict(int),
  EPCT.ALERTS         : 0,
  EPCT.LAST_ALERT_TS  : 0,
  EPCT.FIRST_SEEN     : None,    
  EPCT.LAST_SEEN      : None,
  
  EPCT.LAST_EPOCH_RESTARTS : [], # this will not function without a save-reload mechanism
  
  EPCT.CURRENT_EPOCH  : {
    EPCT.ID               : None,
    EPCT.HB_TIMESTAMPS   : set(),
  },
  
  EPCT.LAST_EPOCH : {
    EPCT.ID : None,
    EPCT.HB_TIMESTAMPS : set(),
  }
}

def str2date(date_str):
  return datetime.strptime(date_str, "%Y-%m-%d %H:%M:%S")

def _get_node_template(name):
  data = deepcopy(_NODE_TEMPLATE)
  data[EPCT.NAME] = name
  return data


class EpochsManager(Singleton):
  
  def build(self, owner, debug_date=None, debug=None):
    """

    """
    if debug is None:
      debug = EPOCH_MANAGER_DEBUG

    self.__last_state_log = 0

    self._epoch_era_setup()
    
    self.owner = owner
    self.__current_epoch = None
    self.__data = {}
    self.__full_data = {}
    self.__eth_to_node = {}
    try:
      debug = int(debug)
    except Exception as e:
      self.P("Error setting debug: {}".format(e), color='r')
      debug = 1
    self.__debug = debug
    self._set_dbg_date(debug_date)

    loaded = self._load_status()
    self.maybe_close_epoch()

    self.P(
      "EpochsMgr v{}, dbg:{}, epoch #{}, GENESIS=[{}] Int/Ep: {}, Sec/Int: {} ".format(
        EPOCH_MANAGER_VERSION, self.__debug, 
        self.get_current_epoch(), self.__genesis_date_str,
        self.__epoch_intervals, self.__epoch_interval_seconds,
      ),
      color='m',
      boxed=True
    )
    return

  @property
  def data(self):
    return self.__data
  
  @property
  def full_data(self):
    return self.__full_data


  @property
  def genesis_date(self):
    return self.__genesis_date
  
  @property
  def epoch_length(self):
    return self.__epoch_intervals * self.__epoch_interval_seconds


  def _epoch_era_setup(self):    
    try:
      self.__epoch_intervals = int(os.environ.get(
        ct.BASE_CT.EE_EPOCH_INTERVALS_KEY, ct.DEFAULT_EPOCH_INTERVALS
      ))
      if ct.BASE_CT.EE_EPOCH_INTERVALS_KEY in os.environ:
        self.P("Epoch intervals set from ENV: {}".format(self.__epoch_intervals), color='m')
      else:
        self.P("Epoch intervals set from default: {}".format(self.__epoch_intervals), color='m')   
    except Exception as e:
      self.P("Error setting epoch intervals: {}. Defaulting.".format(e), color='r')
      self.__epoch_intervals = ct.DEFAULT_EPOCH_INTERVALS
      
    try:
      self.__epoch_interval_seconds = int(os.environ.get(
        ct.BASE_CT.EE_EPOCH_INTERVAL_SECONDS_KEY, ct.DEFAULT_EPOCH_INTERVAL_SECONDS
      ))
      if ct.BASE_CT.EE_EPOCH_INTERVAL_SECONDS_KEY in os.environ:
        self.P("Epoch interval seconds set from ENV: {}".format(self.__epoch_interval_seconds), color='m')
      else:
        self.P("Epoch interval seconds set from default: {}".format(self.__epoch_interval_seconds), color='m')
    except Exception as e:
      self.P("Error setting epoch interval seconds: {}. Defaulting.".format(e), color='r')
      self.__epoch_interval_seconds = ct.DEFAULT_EPOCH_INTERVAL_SECONDS
    
    # for Genesis epoch date is fair to use .replace(utc) in order to have a timezone aware date
    # and not consider the local timezone
    try:
      genesis_epoch_date_env = str(os.environ.get(ct.EE_GENESIS_EPOCH_DATE_KEY, ct.DEFAULT_GENESYS_EPOCH_DATE))
      if len(genesis_epoch_date_env) != len(ct.DEFAULT_GENESYS_EPOCH_DATE):
        genesis_epoch_date_env = ct.DEFAULT_GENESYS_EPOCH_DATE
      if ct.EE_GENESIS_EPOCH_DATE_KEY in os.environ:
        self.P("Genesis epoch date read from ENV: {}".format(genesis_epoch_date_env), color='m')
      else:
        self.P("Genesis epoch date set from default: {}".format(genesis_epoch_date_env), color='m')
    except Exception as e:
      self.P("Error setting genesis epoch date: {}. Defaulting to {}".format(e, ct.DEFAULT_GENESYS_EPOCH_DATE), color='r')
      genesis_epoch_date_env = ct.DEFAULT_GENESYS_EPOCH_DATE
    self.__genesis_date_str = genesis_epoch_date_env
    self.__genesis_date = self.log.str_to_date(self.__genesis_date_str).replace(tzinfo=timezone.utc)
    
    self.__node_alert_interval = self.__epoch_interval_seconds    
    
    _FULL_DATA_TEMPLATE_EXTRA[ct.EE_GENESIS_EPOCH_DATE_KEY] = self.__genesis_date_str
    _FULL_DATA_TEMPLATE_EXTRA[ct.BASE_CT.EE_EPOCH_INTERVALS_KEY] = self.__epoch_intervals
    _FULL_DATA_TEMPLATE_EXTRA[ct.BASE_CT.EE_EPOCH_INTERVAL_SECONDS_KEY] = self.__epoch_interval_seconds
    
    return


  def _set_dbg_date(self, debug_date):
    """
    Is a str is given the date is assumed to be UTC based.
    """
    if debug_date is not None:
      if isinstance(debug_date, str):
        # this is correct and you are supposed to use a UTC based date string
        debug_date = self.log.str_to_date(debug_date).replace(tzinfo=timezone.utc)
    self._debug_date = debug_date
    return


  def P(self, msg, **kwargs):
    self.log.P('[EPM] ' + msg, **kwargs)
    return


  def start_timer(self, name):
    self.log.start_timer(name, section='epoch')
    return
  
  def stop_timer(self, name):
    self.log.stop_timer(name, section='epoch')
    return
  
  def __compute_eth_to_internal(self):
    if not hasattr(self.owner, "node_address_to_eth_address"):
      return
    for node_addr in self.__data:
      eth_node_addr = self.owner.node_address_to_eth_address(node_addr)
      self.__eth_to_node[eth_node_addr] = node_addr
    return
  
  def eth_to_internal(self, eth_node_addr):
    result = self.__eth_to_node.get(eth_node_addr, None)
    if result is None:
      # attempt to recompute
      self.__compute_eth_to_internal()
      result = self.__eth_to_node.get(eth_node_addr, None)
    return result
  
  def get_node_name(self, node_addr):
    """ 
    Given a node address, returns the name of the node.
    """
    return self.owner.network_node_eeid(node_addr)
  
  def __get_max_hb_per_epoch(self):
    max_hb = 0
    addr = self.owner.node_addr
    eeid = self.owner.node_name
    interval = self.owner.network_node_hb_interval(addr=addr)
    if interval is None:
      raise ValueError("Heartbeat interval not found for node: {} ({})".format(addr, eeid))
    nr_hb = 24 * 3600 // interval
    return nr_hb
  
  
  def __debug_status(self):
    if self.__debug:
      self.get_stats(display=True)
    #endif debug
    return
  
  
  def __trim_history(self, trimmable_type=(list, tuple, deque)):
    for full_data_key in _FULL_DATA_TEMPLATE_EXTRA:
      data = self.__full_data[full_data_key]
      is_trimable = isinstance(data, trimmable_type)
      if is_trimable:
        data_len = len(data)
        if data_len > SYNC_HISTORY_SIZE:
          self.__full_data[full_data_key] = self.__full_data[full_data_key][-SYNC_HISTORY_SIZE:]
      else:
        if full_data_key not in self.__full_data:
          self.__full_data[full_data_key] = _FULL_DATA_TEMPLATE_EXTRA[full_data_key] 
    return


  def save_status(self):
    with self.log.managed_lock_resource(EPOCHMON_MUTEX):
      self.__save_status()
    return

      
  def __save_status(self):
    """
    Saves the epochs status to disk called ONLY by `maybe_close_epoch` method - uses the critical section of the maybe_close_epoch method.
    If used separately, make sure to use a lock.
    """
    self.P(f"{self.__class__.__name__} saving epochs status for {len(self.__data)} nodes...")
    
    self.__full_data[SYNC_SAVES_TS].append(self.date_to_str())
    self.__full_data[SYNC_SAVES_EP].append(self.__current_epoch)
    self.__trim_history()
    _full_data_copy = deepcopy(self.__full_data) # maybe not needed

    self.log.save_pickle_to_data(
      data=_full_data_copy, 
      fn=FN_NAME,
      subfolder_path=FN_SUBFOLDER,
    )
    return
  
  
  def _load_status(self):
    """
    NOTE: 2025-01-23 / AID: 
    ----------------------------
    This method is called only once at the beginning of the class initialization and it will
    load the data previously saved to disk (via `__save_status` method) that in turn is called
    by `maybe_close_epoch` method. Thus the status is saved only when the epoch changes (and the
    hb timestamps are reset).
    This means that if a restart is done during a epoch, the data will be loaded from the last 
    reset resulting in a loss of data for the current epoch and the invalidation of the node
    capacity to act as a validator for the current epoch. This is a security feature to prevent
    fraud.
    ------------------------------------------------
    HOWEVER for TRUSTED nodes a procedure of save-reload should be implemented to ensure the
    data is not lost in case of a restart during an epoch but rather preserved and reloaded.
        
    """
    result = False
    exists = self.log.get_data_file(FN_FULL) is not None
    last_epoch_save = None
    if exists:
      self.P("Previous epochs state found. Current oracle era specs:\n{}".format(
        json.dumps(self.get_era_specs(), indent=2)
      ))
      _full_data = self.log.load_pickle_from_data(
        fn=FN_NAME,
        subfolder_path=FN_SUBFOLDER
      )
      if _full_data is not None:
        missing_fields = False
        try:
          # This is retrieved in order to close the epoch in which the last saved happened if needed.
          epochs_of_last_saves = _full_data.get(SYNC_SAVES_EP) or []
          if len(epochs_of_last_saves) > 0:
            last_epoch_save = epochs_of_last_saves[-1]
          # endif epochs of last saves
          dct_to_display = {k: v for k, v in _full_data.items() if k not in [SYNC_NODES, SYNC_SIGNATURES]}
          self.P("Loaded epochs status with {} (current={}) nodes and specs:\n{}".format(
            len(_full_data.get(SYNC_NODES, [])),len(self.__data), json.dumps(dct_to_display, indent=2)
          ))
          for field in _FULL_DATA_MANDATORY_FIELDS:
            if field not in _full_data:
              missing_fields = True
              self.P(f"Missing mandatory field: {field}", color='r')
            # endif field not present
          # endfor mandatory fields
        except Exception as e:
          self.P(f"Error loading epochs status: {e}\n", color='r')
          missing_fields = True
        # end try-except
        if missing_fields:
          # old format
          self.P("Attempting to load old epochs status format. Dropping data", color='r')
          self.__full_data = {
            SYNC_NODES : self.__data,
            SYNC_LAST_EPOCH : INITIAL_SYNC_EPOCH,
          }
        else:
          # new format
          loaded_genesis_date = _full_data.get(ct.EE_GENESIS_EPOCH_DATE_KEY, self.__genesis_date_str)
          loaded_intervals = _full_data.get(ct.BASE_CT.EE_EPOCH_INTERVALS_KEY, self.__epoch_intervals)
          loaded_interval_seconds = _full_data.get(ct.BASE_CT.EE_EPOCH_INTERVAL_SECONDS_KEY, self.__epoch_interval_seconds)
          if (
            loaded_genesis_date != self.__genesis_date_str or
            loaded_intervals != self.__epoch_intervals or
            loaded_interval_seconds != self.__epoch_interval_seconds
          ):
            self.P(
              f"Wrong epoch conf: {loaded_genesis_date}, {loaded_intervals}, {loaded_interval_seconds} vs {self.__genesis_date_str}, {self.__epoch_intervals}, {self.__epoch_interval_seconds}", 
              color='r', boxed=True,
            )
          else:
            # loaded data is full data
            self.__full_data = _full_data
            self.__data = _full_data[SYNC_NODES]
        # end if using new format

        # This is for the third version of the format and is done in this way
        # for maintaining backward compatibility
        # The only difference between second and third format is the following:
        # - the second format had a list of signatures for each availability value from every epoch of every node
        # - the third format has a dictionary of signatures for each epoch aggregated availabilities, thus the data
        # is no longer in the individual node data but in the full data
        self.__full_data[SYNC_SIGNATURES] = self.__full_data.get(SYNC_SIGNATURES, defaultdict(dict))
        self.__full_data[SYNC_AGREEMENT_CID] = self.__full_data.get(SYNC_AGREEMENT_CID, {})
        for node_address, node_data in self.__data.items():
          if EPCT.SIGNATURES in node_data:
            node_data.pop(EPCT.SIGNATURES)
          # endif node_data contains key from old format
        # endfor node data
        result = True
      else:
        self.P("Error loading epochs status.", color='r')
    else:
      self.P(f"No previous epochs status found in {FN_FULL}.", color='r')

    # Retrieved in case of a restart right before the epoch change.
    self.__current_epoch = last_epoch_save
    self.__add_empty_fields()
    self.__compute_eth_to_internal()
    if result:
      self.__full_data[SYNC_RELOADS].append(self.date_to_str())
      self.P(f"Epochs status loaded with {len(self.__data)} nodes", boxed=True)
    #endif exists
    # TODO: further review if this can call maybe_close_epoch
    #  At first inspection it seems there is no issue.
    self.__debug_status()
    return result

  def __add_empty_fields(self):
    """
    Use this method to add missing fields to the loaded data structure.

    """
      
    template = deepcopy(_NODE_TEMPLATE)
    for node_addr in self.__data:
      for key in template:
        if key not in self.__data[node_addr]:
          self.__data[node_addr][key] = template[key]

    if SYNC_NODES not in self.__full_data:
      self.__full_data[SYNC_NODES] = self.__data
          
    template2 = deepcopy(_FULL_DATA_TEMPLATE_EXTRA)  # here we load the epoch specs
    for full_data_key in template2:
      if full_data_key not in self.__full_data:
        self.__full_data[full_data_key] = template2[full_data_key]
    return

  def get_epoch_id(self, date : any):
    """
    Given a date as string or datetime, returns the epoch id - ie the number of days since 
    the genesis epoch.

    Parameters
    ----------
    date : str or date
      The date as string that will be converted to epoch id.
    """
    if isinstance(date, str):
      # remove milliseconds from string
      date = date.split('.')[0]
      date = self.log.str_to_date(date)
      # again this is correct to replace in order to have a timezone aware date
      # and not consider the local timezone. the `date` string naive should be UTC offsetted
      date = date.replace(tzinfo=timezone.utc) 
    # compute difference between date and self.__genesis_date in seconds
    elapsed_seconds = (date - self.__genesis_date).total_seconds()
    
    # the epoch id starts from 0 - the genesis epoch
    # the epoch id is the number of days since the genesis epoch
    # # TODO: change this if we move to start-from-one offset by adding +1
    # OBS: epoch always ends at AB:CD:59 no matter what 
    epoch_id = int(elapsed_seconds / self.epoch_length) 
    return epoch_id
  
  def epoch_to_date(self, epoch_id=None):
    """
    Given an epoch id, returns the date as string.

    Parameters
    ----------
    epoch_id : int
      the epoch id
    """
    if epoch_id is None:
      epoch_id = self.get_time_epoch()
    # TODO: change this if we move to start-from-one offset with (epoch_id - 1)
    date = self.__genesis_date + timedelta(seconds=(epoch_id * self.epoch_length))
    str_date = datetime.strftime(date, format="%Y-%m-%d %H:%M:%S")
    return str_date
  
  def date_to_str(self, date : datetime = None, move_to_utc : bool = False):
    """
    Converts a date to string.
    """
    if date is None:
      date = self.get_current_date()
    if move_to_utc:
      # as you pass a date with timezone info, the conversion to UTC is done by astimezone
      # and then the date is converted to string
      date = date.astimezone(timezone.utc)
    return datetime.strftime(date, format=ct.HB.TIMESTAMP_FORMAT_SHORT)
  
    
  
  def get_current_date(self):
    if self._debug_date is not None:
      return self._debug_date
    else:
      # we convert local time to UTC time
      return datetime.now(timezone.utc)
        
  def get_time_epoch(self):
    """
    Returns the current epoch id.
    """
    return self.get_epoch_id(self.get_current_date())
  
  def get_current_epoch(self):
    """
    Returns the current epoch id using `get_time_epoch`.
    """
    return self.get_time_epoch()
  
  
  def get_hb_utc(self, hb):
    """
    Generates a datetime object from a heartbeat and returns the UTC datetime.
    
    The algorithm is as follows:
    - get the remote timestamp from the heartbeat
    - get the remote timezone from the heartbeat
    - convert the remote timestamp to a datetime object
    - convert the remote datetime to UTC datetime by subtracting the offset hours
    - return the UTC datetime
    
    TODO:
    - add a check for the timezone format
    - add a check for the timestamp format    
    

    Parameters
    ----------
    hb : dict
      the hb object

    Returns
    -------
    datetime.datetime
    """
    ts = hb[ct.PAYLOAD_DATA.EE_TIMESTAMP]
    tz = hb.get(ct.PAYLOAD_DATA.EE_TIMEZONE, "UTC+0")        
    remote_datetime = datetime.strptime(ts, ct.HB.TIMESTAMP_FORMAT)
    offset_hours = int(tz.replace("UTC", ""))
    utc_datetime = remote_datetime - timedelta(hours=offset_hours)
    # the utc_datetime is naive so we need to add the timezone info
    utc_datetime = utc_datetime.replace(tzinfo=timezone.utc)
    return utc_datetime
  
  
  
  def __reset_timestamps(self, node_addr):
    """
    Resets the current epoch timestamps for a node.

    Parameters
    ----------
    node_addr : str
      The node address.
    """
    self.__data[node_addr][EPCT.LAST_EPOCH] = deepcopy(self.__data[node_addr][EPCT.CURRENT_EPOCH])
    self.__data[node_addr][EPCT.CURRENT_EPOCH][EPCT.HB_TIMESTAMPS] = set()
    self.__data[node_addr][EPCT.CURRENT_EPOCH][EPCT.ID] = self.get_time_epoch()
    return


  def __reset_all_timestamps(self):
    for node_addr in self.__data:
      self.__reset_timestamps(node_addr)
    return
  
  # FIXME: this method does not work as expected
  def __calculate_avail_seconds(self, timestamps, time_between_heartbeats=10):
    """
    This method calculates the availability of a node in the current epoch 
    based on the timestamps.

    Parameters
    ----------
    timestamps : set
      The set of timestamps for the current epoch.

    time_between_heartbeats: int
      Mandatory time between heartbeats in seconds.
    
    Returns
    -------
    int
      The availability seconds interval.
    """
    avail_seconds = 0
    nr_timestamps = len(timestamps)
    
    # need at least 2 hb timestamps to compute an interval 
    if nr_timestamps <= 1:
      return 0

    start_timestamp = timestamps[0]
    end_timestamp = timestamps[0]
    for i in range(1, nr_timestamps):
      # timestams should and must be sorted and in the same epoch
      delta = (timestamps[i] - timestamps[i - 1]).seconds
      # the delta between timestamps is bigger than the max heartbeat interval
      # or less than half the heartbeat interval (ignore same heartbeat)
      # TODO(AID): how can a heartbeat be sent more than once?
      # TODO: detect fraud mechanism (someone spams with heartbeats)
      if delta > (time_between_heartbeats + 5) or delta < (time_between_heartbeats / 2):
        # this gets triggered when the delta is too big or too small so last interval 
        # is considered invalid thus we compute up-to-last-valid interval availability
        # (ended with the last set of end_timestamp as end of interval
        avail_seconds += (end_timestamp - start_timestamp).seconds
        start_timestamp = timestamps[i]
      # endif delta

      # change the end of the current interval
      end_timestamp = timestamps[i]
    #endfor each hb timestamp

    # add the last interval length
    avail_seconds += (end_timestamp - start_timestamp).seconds
    return avail_seconds    


  def __calc_node_avail_seconds(self, node_addr, time_between_heartbeats=10, return_timestamps=False):
    if node_addr not in self.__data:
      self.__initialize_new_node(node_addr)
    # endif

    node_data = self.__data[node_addr]
    current_epoch_data = node_data[EPCT.CURRENT_EPOCH]
    timestamps = current_epoch_data[EPCT.HB_TIMESTAMPS]
    current_epoch = current_epoch_data[EPCT.ID]
    lst_timestamps = sorted(list(timestamps))
    avail_seconds = self.__calculate_avail_seconds(
      lst_timestamps, time_between_heartbeats=time_between_heartbeats
    )
    if return_timestamps:
      return avail_seconds, lst_timestamps, current_epoch
    return avail_seconds
    
  def get_current_epoch_availability(self, node_addr=None, time_between_heartbeats=10):
    # TODO: change this if we move to start-from-one offset
    epoch_start = self.__genesis_date + timedelta(
      seconds=(self.epoch_length * self.get_time_epoch()) # -1 if 1st epoch is genesis + length
    )
    max_possible_from_epoch_start = (self.get_current_date() - epoch_start).seconds

    if node_addr is None:
      node_addr = self.owner.node_addr
    # if node not seen yet, return None
    if node_addr not in self.__data:
      return None
    
    avail_seconds = self.__calc_node_avail_seconds(
      node_addr, 
      time_between_heartbeats=time_between_heartbeats
    )
    if max_possible_from_epoch_start == 0:
      prc_available = 0
    else:
      prc_available = round(avail_seconds / max_possible_from_epoch_start, 4)
    return prc_available

  def get_current_epoch_start(self):
    """
    Returns the start date of the current epoch.

    The start date is computed as the genesis date plus the epoch length multiplied by the current epoch id.
    """
    return self.__genesis_date + timedelta(seconds=(self.epoch_length * self.get_time_epoch()))

  def get_current_epoch_end(self):
    """
    Returns the end date of the current epoch.

    The end date is computed as the start date of the current epoch plus the epoch length.
    """
    return self.get_current_epoch_start() + timedelta(seconds=self.epoch_length)

  def __recalculate_current_epoch_for_node(self, node_addr, time_between_heartbeats=10):
    """
    This method recalculates the current epoch availability for a node. 
    It should be used when the epoch changes just before resetting the timestamps.

    Parameters
    ----------
    node_addr : str
      The node address.
    """
    avail_seconds, lst_timestamps, current_epoch = self.__calc_node_avail_seconds(
      node_addr, time_between_heartbeats=time_between_heartbeats,
      return_timestamps=True
    )
    max_possible = self.epoch_length
    prc_available = round(avail_seconds / max_possible, 4) # DO NOT USE 100% but 1.0 
    record_value = round(prc_available * EPOCH_MAX_VALUE)
    self.__data[node_addr][EPCT.EPOCHS][current_epoch] = record_value
    
    if self.__debug:
      try:
        node_name = self.__data[node_addr][EPCT.NAME]
        node_name = node_name[:8]
        start_date, end_date = None, None
        if len(lst_timestamps) >= 1:
          start_date = self.date_to_str(lst_timestamps[0])
          end_date = self.date_to_str(lst_timestamps[-1])
        str_node_addr = node_addr[:8] + '...' + node_addr[-3:]
        self.P("{:<8}<{}> avail in ep {}: {} ({:.2f}%) from {} to {}".format(
          node_name, str_node_addr, current_epoch, 
          record_value, prc_available * 100, start_date, end_date
        ))
      except Exception as e:
        self.P("Error calculating availability for node: {}".format(node_addr), color='r')
        self.P(str(e), color='r')
    return prc_available, current_epoch


  def __recalculate_current_epoch_for_all(self):
    """
    This method recalculates the current epoch availability for all nodes using the recorded 
    timestamps.
    
    NOTE: this method should be called after the epoch has changed and the timestamps have been reset 
    within a critical section (mutex) as already done in `maybe_close_epoch`.
    """    
    self.P("Recalculating epoch {} availability for all nodes during epoch {}...".format(
      self.__current_epoch, self.get_time_epoch()
    ))

    # if current node was not 100% available, do not compute availability for other nodes
    self.start_timer('recalc_node_epoch')
    available_prc, current_epoch = self.__recalculate_current_epoch_for_node(
      self.owner.node_addr
    )
    self.stop_timer('recalc_node_epoch')
    # get the record value for the current node is actually redundant
    record_value = self.__data[self.owner.node_addr][EPCT.EPOCHS][current_epoch]
    
    # we can use available_prc or record_value to check if the current node >= SUPERVISOR_MIN_AVAIL
    # prc = available_prc is the same as record_value / EPOCH_MAX_VALUE
    prc = round(record_value / EPOCH_MAX_VALUE, 4) 
    was_up_throughout_current_epoch = prc >= ct.SUPERVISOR_MIN_AVAIL_PRC

    if not was_up_throughout_current_epoch:
      msg = "Current node was {:.2f}% < {:.0f}%, available in epoch {} and so cannot compute " \
            "availability scores for other nodes".format(
              prc * 100, ct.SUPERVISOR_MIN_AVAIL_PRC * 100, current_epoch
            )
      self.P(msg, color='r')
    else:
      self.start_timer('recalc_all_nodes_epoch')
      for node_addr in self.__data:
        self.start_timer('recalc_node_epoch')
        self.__recalculate_current_epoch_for_node(node_addr)
        self.stop_timer('recalc_node_epoch')
      self.stop_timer('recalc_all_nodes_epoch')
    # endif current node was not 100% available
    return


  def maybe_close_epoch(self):
    """
    This method checks if the current epoch has changed and if so, it closes the current epoch and 
    starts a new one. Closing the epoch implies recalculating the current epoch node availability 
    for all nodes and then resetting the timestamps.
    """
    result = 0 # assume no epoch change
    with self.log.managed_lock_resource(EPOCHMON_MUTEX):
      current_epoch = self.get_time_epoch()
      if self.__current_epoch is None:
        self.__current_epoch = current_epoch
        self.P("Starting epoch: {}".format(self.__current_epoch))
      elif current_epoch != self.__current_epoch:
        if current_epoch != (self.__current_epoch + 1):
          self.P("Epoch jump detected. Current epoch {} vs Last epoch {}".format(
            current_epoch, self.__current_epoch), color='r'
          )
        self.P("Closing epoch {} at start of epoch {}".format(self.__current_epoch, current_epoch))
        result = self.__current_epoch
        self.__recalculate_current_epoch_for_all()
        self.P("Starting epoch: {}".format(current_epoch))
        self.__current_epoch = current_epoch 
        self.__reset_all_timestamps()
        self.__save_status()  # save fresh status current epoch
        #endif epoch is not the same as the current one
      #endif current epoch is not None
    return result


  def get_epoch_of_licensing(self, node_addr):
    """
    Returns the epoch in which a node was associated with its current license.
    Parameters
    ----------
    node_addr : str
      The node address.

    Returns
    -------
    int
      The epoch id.
    """
    # TODO: implement this
    return 1


  def __initialize_new_node(self, node_addr):
    name = self.get_node_name(node_addr)
    name = name[:8]
    node_name = self.get_node_name(node_addr)
    self.__data[node_addr] = _get_node_template(node_name)
    self.__reset_timestamps(node_addr)
    eth_node_addr = self.owner.node_address_to_eth_address(node_addr)
    self.__eth_to_node[eth_node_addr] = node_addr
    self.P("New node {:<8} <{}> / <{}> added to db".format(name, node_addr, eth_node_addr))
    return


  def register_data(self, node_addr, hb):
    """
    This method registers a heartbeat for a node in the current epoch.
    
    Parameters
    ----------
    node_addr : str
      The node address.
      
    hb : dict
      The heartbeat dict.
      
    """
    self.maybe_close_epoch()

    local_epoch = self.get_time_epoch()   
    # maybe first epoch for node_addr
    if node_addr not in self.__data:
      self.__initialize_new_node(node_addr)
    #endif node not in data
    dt_remote_utc = self.get_hb_utc(hb)
    str_date = self.date_to_str(dt_remote_utc)
    if self.__data[node_addr][EPCT.FIRST_SEEN] is None:
      self.__data[node_addr][EPCT.FIRST_SEEN] = str_date
    # check if the hb epoch is the same as the current one
    remote_epoch = self.get_epoch_id(dt_remote_utc)     
    if remote_epoch == local_epoch:
      # the remote epoch is the same as the local epoch so we can register the heartbeat
      with self.log.managed_lock_resource(EPOCHMON_MUTEX):
        # add the heartbeat timestamp for the current epoch
        self.__data[node_addr][EPCT.CURRENT_EPOCH][EPCT.HB_TIMESTAMPS].add(dt_remote_utc)
        self.__data[node_addr][EPCT.LAST_SEEN] = str_date
      # endwith lock
    else:
      self.P("Received invalid epoch {} from node {} on epoch {}".format(
        remote_epoch, node_addr, local_epoch
      ))
    #endif remote epoch is the same as the local epoch
    return
  
  
  def get_node_list(self):
    """
    Returns the list of nodes.
    """
    return list(self.data.keys())
  
  
  def get_node_state(self, node_addr):
    """
    Returns the state of a node in the current epoch.

    Parameters
    ----------
    node_addr : str
      The node address.
    """
    if node_addr not in self.__data:
      return None
    return deepcopy(self.__data[node_addr])
  
  
  def get_node_epochs(self, node_addr, autocomplete=True, as_list=False):
    """
    Returns the epochs availability for a node.

    Parameters
    ----------
    node_addr : str
      The node address.
      
    autocomplete : bool
      If True, the epochs are completed with 0 for missing epochs. Defaults to True in order to ensure continuity in the epochs data.
      
    as_list : bool
      If True, the epochs are returned as a list.
    """
    # TODO: Maybe take into consideration the following use-case:
    #  node A was never seen by the serving oracle, but was licensed before the last
    #  faulty epoch.
    if node_addr not in self.__data:
      return None
    dct_state = self.get_node_state(node_addr)
    dct_epochs = dct_state[EPCT.EPOCHS]
    current_epoch = self.get_time_epoch()
    epochs = list(range(1, current_epoch))
    if autocomplete or as_list:
      for epoch in epochs:
        if epoch not in dct_epochs:
          dct_epochs[epoch] = 0
    # endif autocomplete

    # Apply faulty epochs if any.
    # In case consensus was not achieved from various reasons after the ending of an epoch
    # all nodes that were licensed during or prior to that epoch will be considered as fully available.
    lst_faulty_epochs = self.get_faulty_epochs()
    epoch_of_licensing = self.get_epoch_of_licensing(node_addr=node_addr)
    lst_faulty_epochs = [x for x in lst_faulty_epochs if x >= epoch_of_licensing]
    for faulty_epoch in lst_faulty_epochs:
      dct_epochs[faulty_epoch] = EPOCH_MAX_VALUE
    # endfor faulty epochs

    lst_result = [dct_epochs.get(x, 0) for x in epochs]
    last_epochs = epochs[-5:]
    dct_last_epochs = {x : dct_epochs.get(x, 0) for x in last_epochs}
    non_zero = sum([1 for x in lst_result if x > 0])
    if self.__debug > 1:
      self.P("get_node_epochs({}), {} non zero, last epochs: {}, epoch of licensing: {}, faulty epochs: {}".format(
        node_addr[:10] +'...' + node_addr[-4:], non_zero, str(dct_last_epochs),
        epoch_of_licensing, lst_faulty_epochs
      ))
    if as_list:
      result = lst_result
    else:
      result = dct_epochs
    return result
  
  
  def get_node_last_n_epochs(self, node_addr, n=5, autocomplete=True, as_list=False):
    last_epoch = self.get_time_epoch() - 1
    dct_epochs = self.get_node_epochs(node_addr, autocomplete=autocomplete, as_list=False) or {}
    start = max(1, last_epoch - n + 1)
    lst_epochs = list(range(start, last_epoch + 1))
    result = {x : dct_epochs.get(x, 0) for x in lst_epochs}
    if as_list:
      result = [result[x] for x in lst_epochs]
    return result
  
  
  
  def get_node_epoch(self, node_addr, epoch_id=None, as_percentage=False):
    """
    This method returns the percentage a node was alive in a given epoch.
    The data is returned from already calculated values.

    Parameters
    ----------
    node_addr : str
      The node address.
      
    epoch_id : int
      The epoch id. Defaults to the last epoch

    Returns
    -------
    float
      The value between 0 and 1 representing the percentage of the epoch the node was alive.
    """
    if node_addr not in self.__data:
      return 0
    if epoch_id is None:
      epoch_id = self.get_time_epoch() - 1
    if epoch_id < 1 or epoch_id >= self.get_time_epoch():
      raise ValueError("Invalid epoch requested: {}".format(epoch_id))
    # get the epochs data
    epochs = self.get_node_epochs(node_addr)
    if epochs is None:
      return 0    
    if as_percentage:
      return round(epochs[epoch_id] / EPOCH_MAX_VALUE, 4)
    return epochs[epoch_id]


  def get_node_previous_epoch(self, node_addr, as_percentage=False):
    """
    Returns the last epoch the node was alive.

    Parameters
    ----------
    node_addr : str
      The node address.
    """
    if node_addr not in self.__data:
      return 0
    last_epoch = self.get_time_epoch() - 1
    return self.get_node_epoch(node_addr, epoch_id=last_epoch, as_percentage=as_percentage)

  
  def get_node_last_epoch(self, node_addr, as_percentage=False):
    """
    Alias for get_node_previous_epoch.
    """
    return self.get_node_previous_epoch(node_addr, as_percentage=as_percentage)  


  def get_self_supervisor_capacity(self, as_float=False, start_epoch=None, end_epoch=None):
    """
    Returns the supervisor capacity for all the epochs
    
    Parameters
    ----------
    
    as_float : bool
      If True, the values are returned as floats. If False, the values are returned as bools
      based on (epochs[epoch] >= SUPERVISOR_MIN_AVAIL_UINT8).
      
    start_epoch : int
      The start epoch. Defaults to 1.
      
    end_epoch : int
      The end epoch. Defaults to the current epoch - 1.
      
    
    """
    epochs = self.get_node_epochs(self.owner.node_addr) or defaultdict(int)
    
    start_epoch = start_epoch if isinstance(start_epoch, int) else 1
    end_epoch = end_epoch if isinstance(end_epoch, int) else self.get_time_epoch() - 1
    
    lst_epochs = list(range(start_epoch, end_epoch + 1))

    epoch_certainty_scores = {
      # If the epoch availability was obtained through the consensus mechanism, we
      # consider full certainty, otherwise we consider the value obtained from the node.
      epoch: epochs[epoch] if epoch > self.get_last_sync_epoch() else EPOCH_MAX_VALUE
      for epoch in lst_epochs
    }

    result = {
      epoch:
        (certainty_score >= SUPERVISOR_MIN_AVAIL_UINT8) if not as_float else
        (round(certainty_score / EPOCH_MAX_VALUE, 2))
      for epoch, certainty_score in epoch_certainty_scores.items()
    }
    return result
    

  # TODO: maybe replace this with `get_epoch_of_licensing` after implementing
  def get_node_first_epoch(self, node_addr):
    """
    Returns the first epoch the node was seen as alive by the current oracle.

    Parameters
    ----------
    node_addr : str
      The node address.
    """
    if node_addr not in self.__data:
      return -1
    epochs = list(self.get_node_epochs(node_addr).keys())
    min_epoch = min(epochs)
    return min_epoch
  
  

  def get_era_specs(self):
    """
    This function returns in human readable format the era specifications meaning it will return:
    - the current epoch
    - the current date
    - genesis date
    - epoch intervals
    - epoch interval seconds
    """
    dct_result = {
      'current_epoch' : self.get_current_epoch(),
      'current_date' : self.date_to_str(self.get_current_date()),
      'genesis_date' : self.__genesis_date_str,
      'epoch_intervals' : self.__epoch_intervals,
      'epoch_interval_seconds' :  self.__epoch_interval_seconds,
      'epoch_length' : self.epoch_length,
    }
    return dct_result


  def get_oracle_state(
    self, display=False, 
    start_epoch=None, end_epoch=None,
    as_int=False,
  ):
    """
    Returns the server/oracle state.
    """
    dct_result = self.get_era_specs()
    if start_epoch is None:
      start_epoch = max(1, self.get_current_epoch() - 10)
    certainty = self.get_self_supervisor_capacity(
      as_float=True, start_epoch=start_epoch, end_epoch=end_epoch,
    )
    epochs = sorted(list(certainty.keys()))
    certainty_int = {
      x : int(certainty[x] >= ct.SUPERVISOR_MIN_AVAIL_PRC) for x in epochs
    }
    if as_int:
      certainty = certainty_int
    dct_result['manager'] = {
      'ver' : f"{NODE_VERSION} / {CORE_VERSION} / {SDK_VERSION}",
      'certainty' : certainty, 
    }
    dct_result['manager']['valid'] = sum(certainty_int.values()) == len(certainty)
    dct_result['manager']['supervisor_min_avail_prc'] = ct.SUPERVISOR_MIN_AVAIL_PRC
    if self.get_current_epoch() < 20:
      # at the beginning we dump the epochs
      dct_result['manager']['epochs'] = self.get_node_epochs(self.owner.node_addr)
    for extra_key in _FULL_DATA_INFO_KEYS:
      dct_result['manager'][extra_key.lower()] = self.__full_data.get(extra_key, 'N/A')
    if display:
      self.P("Oracle state:\n{}".format(json.dumps(dct_result, indent=2)))
      self.__last_state_log = time()
    return dct_result


  def get_stats(self, display=False, online_only=False):
    """
    Returns the overall statistics for all nodes.
    """
    stats = {'error' : None}
    best_avail = 0
    NR_HIST = 10
    nr_eps = 0   
    try:
      saves = self.__full_data.get(SYNC_SAVES_TS, 'N/A')
      saves_epoch = self.__full_data.get(SYNC_SAVES_EP, 'N/A')
      restarts = self.__full_data.get(SYNC_RESTARTS, 'N/A')
      current_epoch = self.get_current_epoch()
      start_epoch = max(1, current_epoch - NR_HIST)
      certainty = self.get_self_supervisor_capacity(as_float=True, start_epoch=start_epoch)
      oracle_state = self.get_oracle_state()      
      for node_addr in self.data:        
        try:
          is_online = self.owner.network_node_is_online(
            node_addr, dt_now=self.get_current_date()
          )
          node_version = self.owner.network_node_version(node_addr)
          if online_only and not is_online:
              continue
          dt_netmon_last_seen = self.owner.network_node_last_seen(
            node_addr, 
            dt_now=self.get_current_date(),
            as_sec=False
          )
          last_seen_ago = self.owner.network_node_last_seen(
            node_addr, 
            dt_now=self.get_current_date(),
            as_sec=True
          )
          
          mem_total = self.owner.network_node_total_mem(node_addr)
          mem_avail = self.owner.network_node_avail_mem(node_addr)
          cpu_cores = self.owner.network_node_total_cpu_cores(node_addr)
          cpu_cores_avail = self.owner.network_node_avail_cpu_cores(node_addr)
          disk_total = self.owner.network_node_total_disk(node_addr)
          disk_avail = self.owner.network_node_avail_disk(node_addr)
          
          # GPU
          gpu_name = self.owner.network_node_default_gpu_name(node_addr)
          gpu_mem_total = self.owner.network_node_default_gpu_total_mem(node_addr)
          gpu_mem_avail = self.owner.network_node_default_gpu_avail_mem(node_addr)
          gpu_usage = self.owner.network_node_default_gpu_usage(node_addr)
          default_cuda = self.owner.network_node_default_cuda(node_addr, as_int=False)
          gpu_status, gpu_info = None, None # only in debug mode
          
          # DEBUG:
          if True:
            gpu_status = self.owner.network_node_last_gpu_status(node_addr)
            gpu_info = self.owner.network_node_gpu_summary(node_addr)
          # end DEBUG
        except:
          continue
        
        netmon_last_seen = self.date_to_str(dt_netmon_last_seen) if dt_netmon_last_seen is not None else 'N/A'
        node_name = self.get_node_name(node_addr)
        dct_epochs = self.get_node_epochs(node_addr, as_list=False, autocomplete=True)     
        
        # current epoch hb data
        node_current_epoch_nr_hb = None
        node_current_epoch_1st_hb = None
        node_current_epoch_last_hb = None
        node_current_epoch_id = None
        try: 
          node_current_epoch_data = self.data[node_addr][EPCT.CURRENT_EPOCH]
          node_current_epoch_id = node_current_epoch_data[EPCT.ID]
          node_current_epoch_hb_timestamps = node_current_epoch_data[EPCT.HB_TIMESTAMPS]
          node_current_epoch_hb_timestamps = sorted(list(node_current_epoch_hb_timestamps))
          node_current_epoch_1st_hb = node_current_epoch_hb_timestamps[0] if len(node_current_epoch_hb_timestamps) > 0 else None
          node_current_epoch_last_hb = node_current_epoch_hb_timestamps[-1] if len(node_current_epoch_hb_timestamps) > 0 else None
          node_current_epoch_nr_hb = len(node_current_epoch_hb_timestamps)
        except:
          pass
        
        
        # process the previous epoch hb data
        node_last_epoch_data = self.data[node_addr][EPCT.LAST_EPOCH]
        node_last_epoch_id = node_last_epoch_data[EPCT.ID]
        node_last_epoch_hb_timestamps = node_last_epoch_data[EPCT.HB_TIMESTAMPS]
        node_last_epoch_hb_timestamps = sorted(list(node_last_epoch_hb_timestamps))
        node_last_epoch_1st_hb = node_last_epoch_hb_timestamps[0] if len(node_last_epoch_hb_timestamps) > 0 else None
        node_last_epoch_1st_hb = self.date_to_str(node_last_epoch_1st_hb)
        node_last_epoch_last_hb = node_last_epoch_hb_timestamps[-1] if len(node_last_epoch_hb_timestamps) > 0 else None
        node_last_epoch_last_hb = self.date_to_str(node_last_epoch_last_hb)
        node_last_epoch_nr_hb = len(node_last_epoch_hb_timestamps)

        node_last_epoch_avail = self.get_node_previous_epoch(node_addr)

        node_last_epoch_avail_local = round(
          self.__calculate_avail_seconds(node_last_epoch_hb_timestamps) / self.epoch_length, 4
        )
        
        epochs_ids = sorted(list(dct_epochs.keys()))
        epochs = [dct_epochs[x] for x in epochs_ids]
        selection = epochs_ids[-NR_HIST:]
        str_last_epochs = str({x : dct_epochs.get(x, 0) for x in selection})
        str_certainty =  ", ".join([
          f"{x}={'Y' if certainty.get(x, 0) >= ct.SUPERVISOR_MIN_AVAIL_PRC else 'N'}" 
          for x in selection
        ])    
        MAX_AVAIL = EPOCH_MAX_VALUE * len(epochs) # max avail possible for this node
        score = sum(epochs)      
        avail = round(score / (MAX_AVAIL + 1e7), 4)
        best_avail = max(best_avail, avail)
        non_zero = len([x for x in epochs if x > 0])
        nr_eps = len(epochs)
        prev_epoch = self.get_time_epoch() - 1
        first_seen = self.data[node_addr][EPCT.FIRST_SEEN]
        last_seen = self.data[node_addr][EPCT.LAST_SEEN]
        eth_addr = self.owner.node_address_to_eth_address(node_addr)
        if nr_eps != prev_epoch:
          msg = "Epochs mismatch for node: {} - total {} vs prev {}".format(
            node_addr, nr_eps, prev_epoch
          )
          msg += "\nEpochs: {}".format(dct_epochs)
          msg += "\nCurrent epoch: {}".format(current_epoch)
          msg += "\nPrevious epoch: {}".format(self.get_time_epoch() - 1)
          self.P(msg, color='r')
          if abs(nr_eps - prev_epoch) > 1:
            raise ValueError(msg)
        stats[node_addr] = {
          'eth_addr' : eth_addr,
          'alias' : node_name,
          'ver' : node_version,
          'last_state' : netmon_last_seen,
          'last_seen_ago' : self.log.elapsed_to_str(last_seen_ago),
          'non_zero' : non_zero,
          'overall_availability' : avail,
          'score' : score,
          'first_check' : first_seen,
          'last_check' : last_seen,
          
          'resources' : {
            'mem_total' : mem_total,
            'mem_avail' : mem_avail,
            'cpu_cores' : cpu_cores,
            'cpu_cores_avail' : cpu_cores_avail,
            'disk_total' : disk_total,
            'disk_avail' : disk_avail,
            'default_cuda' : default_cuda,
            'gpu_name' : gpu_name,
            'gpu_mem_total' : gpu_mem_total,
            'gpu_mem_avail' : gpu_mem_avail,
            'gpu_usage' : gpu_usage,
            'gpu_status' : gpu_status,
            'gpu_info' : gpu_info,
          },
          
          'current_epoch' : {
            'id' : node_current_epoch_id,
            'nr_hb' : node_current_epoch_nr_hb,
            '1st_hb' : str(node_current_epoch_1st_hb),
            'last_hb' : str(node_current_epoch_last_hb),
          },
                  
          'recent_history' : {
            'last_10_ep' : str_last_epochs,
            'certainty' : str_certainty,
            'last_epoch_id' : node_last_epoch_id,
            'last_epoch_nr_hb' : node_last_epoch_nr_hb,
            'last_epoch_1st_hb' : node_last_epoch_1st_hb,
            'last_epoch_last_hb' : node_last_epoch_last_hb,
            'last_epoch_avail_local' : node_last_epoch_avail_local,
            'last_epoch_avail' : node_last_epoch_avail,
          }
        }
        if node_addr == self.owner.node_addr:
          stats[node_addr]['oracle'] = oracle_state
        #endif node is current node
      #endfor each node
      if display:
        str_stats = json.dumps(stats, indent=2)
        self.P("EpochManager report at ep {} (max_score: {}, nr_eps: {}):\nRecent saves: {}\nRecent saves epochs: {}\nRecent restars: {}\nOracle info:\n{}\n\nStatuses:\n{}".format(
          current_epoch, best_avail, nr_eps,
          saves, saves_epoch, restarts, 
          json.dumps(oracle_state, indent=2),
          str_stats
        ))
    except Exception as e:
      msg = "Error getting EpochManager stats: {}".format(str(e))
      stats['error'] = msg
    return stats
  

### Below area contains the methods for availability resulted from multi-oracle sync

  def get_last_sync_epoch(self):
    """
    Returns the last sync epoch.

    Returns
    -------
    int
      The last sync epoch.
    """
    return self.__full_data.get(SYNC_LAST_EPOCH, INITIAL_SYNC_EPOCH)


  def get_epoch_availability(self, epoch, return_additional=False):
    """
    Returns the availability table for a given epoch.

    Parameters
    ----------
    epoch : int
      The epoch id.

    return_additional: bool
      Whether to return the signatures and CID alongside the availability table or not

    Returns
    -------
    (availability_table, signatures, cid) if return_additional else availability_table, where
    availability_table: dict
      The availability table for the specified epoch.
    signatures: dict
      The signatures recorded for the current epoch.
    cid: str
      The CID of the agreement for the current epoch if it was
      stored in R1FS. None otherwise.
    """

    availability_table = {}
    epoch_signatures = {}
    epoch_agreement_cid = self.__full_data[SYNC_AGREEMENT_CID].get(epoch, None)

    if self.is_epoch_valid(epoch):
      for node_addr in self.__data:
        epochs: defaultdict = self.get_node_epochs(node_addr, as_list=False)
        availability_table[node_addr] = epochs.get(epoch, 0)
      # end for each node
      # self.__data[EPCT.SIGNATURES] is a defaultdict(dict), thus there is no need for .get() here
      epoch_signatures = self.__full_data[SYNC_SIGNATURES][epoch]
    # endif epoch is valid

    return (availability_table, epoch_signatures, epoch_agreement_cid) if return_additional else availability_table


  def get_faulty_epochs(self):
    """
    Get all passed epochs for which the consensus could not be achieved.
    Returns
    -------
    list : all passed epochs for which the consensus could not be achieved.
    """
    return self.__full_data[FAULTY_EPOCHS]


  def is_epoch_faulty(self, epoch):
    """
    Checks if an epoch is marked as faulty.

    Parameters
    ----------
    epoch : int
      The epoch id.

    Returns
    -------
    bool
      True if the epoch is marked as faulty, False otherwise.
    """
    return epoch in self.__full_data[FAULTY_EPOCHS]


  def is_epoch_valid(self, epoch):
    """
    Checks if an epoch is valid(consensus was achieved).

    Parameters
    ----------
    epoch : int
      The epoch id.

    Returns
    -------
    bool
      True if the epoch is valid, False otherwise.
    """
    return not self.is_epoch_faulty(epoch)


  def mark_epoch_as_faulty(self, epoch, debug=True):
    """
    Marks an epoch as faulty. This means that consensus was not achieved for the given epoch.
    In this case all nodes with licenses associated prior to it will be considered as fully available.
    Parameters
    ----------
    epoch : int
      The epoch id.
    debug : bool
      If True, the debug messages are displayed.

    Returns
    -------
    bool
      True if the epoch was successfully marked as faulty, False otherwise.
    """
    success = True
    last_sync_epoch = self.get_last_sync_epoch()

    if epoch <= last_sync_epoch:
      if debug:
        self.P(f"Epoch {epoch} is not greater than last sync epoch {last_sync_epoch}. Skipping marking.", color='r')
      success = False
      return success

    if epoch in self.__full_data[FAULTY_EPOCHS]:
      if debug:
        self.P(f"Epoch {epoch} is already marked as faulty. Skipping marking.", color='r')
      success = False
      return success

    self.__full_data[FAULTY_EPOCHS].append(epoch)
    if debug:
      self.P(f"Epoch {epoch} marked as faulty.")
    return success


  def unmark_epoch_as_faulty(self, epoch, debug=True):
    """
    Unmarks an epoch as faulty. This means that consensus was achieved for the given epoch.
    This can happen if the epoch is marked as faulty, but later it's requested again and
    a valid consensus was, in fact, reached.
    Will be used only in `update_epoch_availability` method.

    Parameters
    ----------
    epoch : int
      The epoch id.

    Returns
    -------
    bool
      True if the epoch was successfully unmarked as faulty, False otherwise.
    """
    success = True
    # TODO: maybe check if the epoch is lower than the last synced one.
    #  Should not be necessary since it's called only from `update_epoch_availability`
    #  after validating the signatures from oracles.

    if epoch not in self.__full_data[FAULTY_EPOCHS]:
      if debug:
        self.P(f"Epoch {epoch} is not marked as faulty. Skipping unmarking.", color='r')
      success = False
      return success

    self.__full_data[FAULTY_EPOCHS].remove(epoch)
    if debug:
      self.P(f"Epoch {epoch} unmarked as faulty.")
    return success


  def add_cid_for_epoch(self, epoch, agreement_cid, debug=False):
    """
    Adds the agreement CID for the given epoch.

    Parameters
    ----------
    epoch : int
      The epoch id.

    agreement_cid : str
      The agreement CID.

    Returns
    -------
    bool
      True if the agreement CID was added successfully, False otherwise.
    """
    signatures = self.__full_data[SYNC_SIGNATURES].get(epoch) or {}
    if len(signatures) == 0:
      if debug:
        self.P(f"No signatures found for epoch {epoch}. Skipping CID addition.", color='r')
      return False
    self.__full_data[SYNC_AGREEMENT_CID][epoch] = agreement_cid
    if debug:
      self.P(f"Agreement CID for epoch {epoch} added successfully.")
    return True


  def update_epoch_availability(
      self, epoch: int, availability_table: dict,
      agreement_signatures: dict, agreement_cid: str,
      debug: bool = False
  ):
    """
    Updates the epoch availability for a given epoch.

    !! IMPORTANT !!
    ---------------
    Make sure the epoch is strictly greater than the last sync epoch.
    It is ideal that this method is called with `epoch == last_sync_epoch + 1`.

    Parameters
    ----------
    epoch : int
      The epoch id.

    availability_table : dict
      The availability table.

    agreement_signatures : dict
      The agreement signatures.

    agreement_cid : str
      The agreement CID in case the agreement was stored in R1FS.

    debug : bool
      If True, the debug messages are displayed.
    """
    success = True
    last_sync_epoch = self.get_last_sync_epoch()

    # assert epoch > last_sync_epoch, \
    #   f"Epoch {epoch} is not greater than last sync epoch {last_sync_epoch}"
    if epoch <= last_sync_epoch:
      self.P(f"Epoch {epoch} is not greater than last sync epoch {last_sync_epoch}. Skipping update.", color='r')
      success = False
      return success

    for node_addr in availability_table:
      if node_addr not in self.__data:
        self.__initialize_new_node(node_addr)
      if debug:
        self.P(f'DEBUG self.__data before update: {self.__data[node_addr]}')
      self.__data[node_addr][EPCT.EPOCHS][epoch] = availability_table[node_addr]
      if debug:
        self.P(f'DEBUG self.__data after update: {self.__data[node_addr]}')
    self.__full_data[SYNC_SIGNATURES][epoch] = agreement_signatures
    self.__full_data[SYNC_AGREEMENT_CID][epoch] = agreement_cid
    self.__full_data[SYNC_LAST_EPOCH] = epoch

    self.P(f"Epoch {epoch} availability updated successfully.")

    if epoch in self.get_faulty_epochs():
      self.P(f"Epoch {epoch} previously marked a faulty. Maybe unmarking...", color='r')
      success = self.unmark_epoch_as_faulty(epoch)
      if success:
        self.P(f"Epoch {epoch} unmarked as faulty.")
    # endif epoch previously marked as faulty

    return success




if __name__ == '__main__':
  from naeural_core.core_logging import Logger
  from naeural_core.main.net_mon import NetworkMonitor
  
  FN_NETWORK = r"_local_cache\_data\network_monitor\db.pkl"
  
  EPOCH_MANAGER_DEBUG = False
  
  l = Logger('EPOCH', base_folder='.', app_folder='_local_cache')
  
  DATES = [
    '2024-07-08 12:00:00',
    '2024-07-07 12:00:00',
    '2024-07-08 12:00:00',
    '2024-07-09 12:00:00',
    '2024-07-10 12:00:00',
  ]
  
  NODES = [
    '0xai_AkyWQ91tdk0QdJfH70nmRG6euFjxwYf1FSC7mBdtIbTh',
    '0xai_AgNxIxNN6RsDqBa0d5l2ZQpy7y-5bnbP55xej4OvcitO',
  ]
  
  # make sure you have a recent (today) save network status
  # eng1 = EpochsManager(log=l, owner=1234, debug_date=DATES[0], debug=True)
  # eng2 = EpochsManager(log=l, owner=None, debug_date=DATES[1])
  # assert id(eng1) == id(eng2)
  
  PREDEFINED_TESTS = {
    'aid_01' : {
      'date' :'2025-01-24 09:07:00',
      'addr' : '0xai_AleLPKqUHV-iPc-76-rUvDkRWW4dFMIGKW1xFVcy65nH'
    },
    'nen-2' : {
      'date' :'2025-01-24 11:26:00',
      'addr' : '0xai_Amfnbt3N-qg2-qGtywZIPQBTVlAnoADVRmSAsdDhlQ-6'      
    }
  }
  
  CHOSEN_TEST = 'nen-2'
  
  CURRENT_DATE = PREDEFINED_TESTS[CHOSEN_TEST]['date']
  NODE_ADDR = PREDEFINED_TESTS[CHOSEN_TEST]['addr']
  
  if True:
    netmon = NetworkMonitor(
      log=l, node_name=CHOSEN_TEST, node_addr=NODE_ADDR,
      # epoch_manager=eng
    )
  else:
    netmon = NetworkMonitor(
      log=l, node_name='aid_hpc', node_addr='0xai_AgNxIxNN6RsDqBa0d5l2ZQpy7y-5bnbP55xej4OvcitO'
    )
    
  # eng.owner = netmon
  eng = netmon.epoch_manager
  
  
  TEST_A = False
  TEST_B = True
  
  if TEST_A:
    assert id(eng) == id(netmon.epoch_manager)  

    has_data = netmon.network_load_status(FN_NETWORK)
    
    if has_data:    
      l.P("Current time epoch is: {} ({})".format(eng.get_time_epoch(), eng.epoch_to_date()))
      
      nodes = netmon.all_nodes
          
      dct_hb = {}
      
      # now check the nodes for some usable data
      _current_epoch = eng.get_time_epoch()
      for node_addr in nodes:
        hbs = netmon.get_box_heartbeats(node_addr)
        idx = -1
        done = False
        good_hbs = defaultdict(list)
        for hb in hbs:
          ep = eng.get_epoch_id(hb[ct.PAYLOAD_DATA.EE_TIMESTAMP])
          if ep >= _current_epoch:
            good_hbs[ep].append(hb)
        if len(good_hbs) > 0:
          dct_hb[node_addr] = good_hbs
      
      l.P("Data available for epochs:\n{}".format(
        "\n".join(["{}: {}".format(x, list(dct_hb[x].keys())) for x in dct_hb]) 
      ))
      
      
      for step in range(5):
        current_date = DATES[step]
        eng._set_dbg_date(current_date)
        epoch = eng.get_epoch_id(current_date)
        l.P("Running step {} - epoch {} / {}".format(
          step, epoch, current_date), color='b'
        )
        epoch_has_data = any([epoch in dct_hb[x] for x in dct_hb])
        if epoch_has_data:
          l.P("Starting registering data for epoch {}...".format(eng.get_current_epoch()), color='b')
        data_counter = 0
        for node_addr in dct_hb:
          for hb in dct_hb[node_addr][epoch]:
            eng.register_data(node_addr, hb)
            data_counter += 1
        if data_counter > 0:
          l.P("Data loaded ({}) for epoch {}.".format(
            data_counter, eng.get_current_epoch()), color='g'
          )
        else:
          l.P("No data registered for epoch {}.".format(eng.get_current_epoch()), color='r')
        #endif had data
      #endfor each step
      final_date = DATES[-1]
      l.P("Done all steps, setting final date: {}".format(final_date), color='b')
      eng._set_dbg_date(final_date)    
      eng.maybe_close_epoch()
      
      l.P('{}: {}'.format(
        eng.get_node_name(NODES[-2]), eng.get_node_epochs(NODES[-2], as_list=True))
      )
      l.P('{}: {}'.format(
        eng.get_node_name(NODES[-1]), eng.get_node_epochs(NODES[-1], as_list=True))
      )    
  #endif TEST_A
  
  if TEST_B:
    str_date = CURRENT_DATE
    debug_date = l.str_to_date(str_date) # get date
    debug_date = debug_date.astimezone(timezone.utc) # convert to UTC
    eng._set_dbg_date(debug_date)
    inf = eng.get_stats(display=True, online_only=True)
    m, t, top = l.get_obj_size(obj=netmon.all_heartbeats, top_consumers=20, return_tree=True)
    # l.P("Heartbeats size: {:,.0f} MB".format(m / 1024 / 1024))
    eng.get_current_epoch_availability(NODE_ADDR)
  


import numpy as np
import cv2

from naeural_core import Logger

if __name__ == '__main__':
  
  l = Logger('CV2T', base_folder='.', app_folder='_cache')
  
  cap = cv2.VideoCapture(0)
  
  
  while(True):
    l.start_timer('cv2_read_ala_bala_portocala_long_name')
    ret, frame = cap.read() 
    l.stop_timer('cv2_read_ala_bala_portocala_long_name')
    if ret:
      cv2.imshow('frame', frame)
      
    if cv2.waitKey(1) & 0xFF == ord('q'):
      break

  rets = []
  rets2 = []
  fails = 0 
  while(True):
    for i in range(100):
      l.start_timer('cv2_grab')
      ret = cap.grab() 
      l.stop_timer('cv2_grab')
      rets.append(ret)
    
    if ret:
      l.start_timer('cv2_retrieve')
      ret2, frame = cap.retrieve()
      l.stop_timer('cv2_retrieve')
      rets2.append(ret2)
      if ret2:
        cv2.imshow('frame', frame)
    else:
      fails += 1
      
    if cv2.waitKey(1) & 0xFF == ord('q'):
      break

  cap.release()
  cv2.destroyAllWindows()
  
  l.show_timers()
  l.P("Fails: {}".format(fails))


class _ThUtilsMixin:
  def scale_coords(self, img1_shape, coords, img0_shape, ratio_pad=None, shrink_factor=None):
    # Rescale coords (xyxy) from img1_shape to img0_shape
    if ratio_pad is None:  # calculate from img0_shape
      gain = min(img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1])  # gain  = old / new
      if shrink_factor is not None:
        gain /= shrink_factor
      pad = (img1_shape[1] - img0_shape[1] * gain) / 2, (img1_shape[0] - img0_shape[0] * gain) / 2  # wh padding
    else:
      gain = ratio_pad[0][0]
      pad = ratio_pad[1]

    coords[:, [x for x in range(coords.shape[1]) if x % 2 == 0]] -= pad[0]  # x padding
    coords[:, [x for x in range(coords.shape[1]) if x % 2 == 1]] -= pad[1]  # y padding
    coords[:, :] /= gain
    self.clip_coords(coords, img0_shape)
    return coords

  def clip_coords(self, boxes, shape):
    # Clip bounding xyxy bounding boxes to image shape (height, width)
    if isinstance(boxes, self.th.Tensor):  # faster individually
      boxes[:, 0].clamp_(0, shape[1])  # x1
      boxes[:, 1].clamp_(0, shape[0])  # y1
      boxes[:, 2].clamp_(0, shape[1])  # x2
      boxes[:, 3].clamp_(0, shape[0])  # y2
    else:  # np.array (faster grouped)
      boxes[:, [x for x in range(boxes.shape[1]) if x % 2 == 0]] = boxes[:, [
          x for x in range(boxes.shape[1]) if x % 2 == 0]].clip(0, shape[1])  # x1, x2
      boxes[:, [x for x in range(boxes.shape[1]) if x % 2 == 1]] = boxes[:, [
          x for x in range(boxes.shape[1]) if x % 2 == 1]].clip(0, shape[0])  # y1, y2



import zipfile
import os


def zipdir(path, ziph):
  # ziph is zipfile handle
  for root, dirs, files in os.walk(path):
    for file in files:
      ziph.write(os.path.join(root, file),
                 os.path.relpath(os.path.join(root, file),
                                 os.path.join(path, '..')))
  return


def get_archive_idx(archive_name, location):
  lst_dir = os.listdir(location)
  lst_filtered = [os.path.splitext(f)[0] for f in lst_dir if archive_name in f and f.endswith('.zip')]
  lst_idxs = [f[len(archive_name):].split('_')[1] for f in lst_filtered]
  lst_idxs_int = [int(f) for f in lst_idxs if f.isnumeric()]
  return max(lst_idxs_int) + 1 if len(lst_idxs_int) > 0 else 0


if __name__ == '__main__':
  dir_path = 'C:/resources/dummy_ds'
  # dir_path = 'C:/resources/empty_ds'
  archive_name = 'dummy_zip'
  location = 'C:/resources'
  it = get_archive_idx(archive_name=archive_name, location=location)
  archive_path = os.path.join(location, f'{archive_name}_{it}.zip')
  zipf = zipfile.ZipFile(archive_path, 'w', zipfile.ZIP_DEFLATED)
  zipdir(dir_path, zipf)
  zipf.close()




import numpy as np
import cv2 


def apply_watermark_from_file(np_bgr, fn_watermark):
  H, W = np_bgr.shape[:2]
  np_watermark = prepare_watermark(
    fn_watermark=fn_watermark,
    target_shape=(H,W)
  )
  return apply_watermark(np_bgr, np_watermark)

def apply_watermark(np_bgr, np_watermark):  
  if isinstance(np_watermark, np.ndarray):
    if np_watermark.shape[:2] != np_bgr.shape[:2]:
      H, W = np_bgr.shape[:2]
      wh, ww = np_watermark.shape[:2]
      nh = H // wh + 1
      nw = W // ww + 1
      w1 = np.tile(np_watermark, (nh, nw, 1))
      np_watermark = w1[:H, :W]
    np_bgr = cv2.addWeighted(np_bgr, 1, np_watermark, 0.3, 0)
  return np_bgr


def prepare_watermark(fn_watermark, target_shape_HW, text="<Execution Engine>"):
  small_watermark = None
  if fn_watermark is not None:
    small_watermark = cv2.imread(fn_watermark)
  H, W = target_shape_HW
  np_watermark = None
  if small_watermark is None:
    small_watermark = np.full((200, 340, 3), fill_value=0, dtype=np.uint8)
    s = text
    org = (15, 190)
    cv2.putText(
      img=small_watermark, 
      text=s, 
      org=org, 
      fontFace=cv2.FONT_HERSHEY_DUPLEX, 
      fontScale=1, 
      color=(255, 255, 255), 
      thickness=2,
    )
    M = cv2.getRotationMatrix2D(org, 30, 1)
    small_watermark = cv2.warpAffine(small_watermark, M, (small_watermark.shape[1], small_watermark.shape[0]))
  wh, ww = small_watermark.shape[:2]
  nh = H // wh + 1
  nw = W // ww + 1
  w1 = np.tile(small_watermark, (nh, nw, 1))
  np_watermark = w1[:H, :W]  
  return np_watermark



import numpy as np
import matplotlib.pyplot as plt

if __name__ == '__main__':
  name_tp = 'TP'
  name_fp = 'FP'
  lst_thr = list(np.arange(0, 1.1, 0.1))
  lst_inf = [
    {
      'TLBR_POS': [1, 2, 3, 4], 
      'PROB_PRC': i/10, 
      'TYPE': 'person',
      'CLASS': 'TP' if i % 2 == 0 else 'FP'
    } for i in range(11)
    ]
  print(lst_inf)
  
  l_sensitivity, l_specificity = [], []
  lst_all_tp = [x for x in lst_inf if x['CLASS'] == name_tp]
  lst_all_fp = [x for x in lst_inf if x['CLASS'] == name_fp]
  for thr in lst_thr:
    l_thr = [x for x in lst_inf if x['PROB_PRC'] >= thr]
    lst_tp_thr = [x for x in l_thr if x['CLASS'] == name_tp]
    lst_fp_thr = [x for x in l_thr if x['CLASS'] == name_fp]
    lst_tn_thr = [x for x in lst_inf if x['PROB_PRC'] < thr and x['CLASS'] == name_tp]
    
    # Sensitivity OR TPR = TP / (TP + FN)
    sensitivity = len(lst_tp_thr) / len(lst_all_tp)
    
    # Specificity = TN / (TN + FP)
    specificity = len(lst_tn_thr) / (len(lst_tn_thr) + len(lst_fp_thr))
    
    l_sensitivity.append(sensitivity)
    l_specificity.append(specificity)
  #endfor
  
  fig = plt.figure()
  plt.plot(l_specificity, l_sensitivity)
  plt.xlabel('Specificity (False Positive Rate)')
  plt.ylabel('Sensitivity (True Positive Rate)')
  plt.show()
  
  
  


import numpy as np
import torch as th

from collections import OrderedDict
import textwrap
import os
from time import time

from naeural_core import DecentrAIObject
from .utils import Pr


_LOSSES = {
  'mse'                 : th.nn.MSELoss,
  'binary_crossentropy' : th.nn.BCELoss,
  'mae'                 : th.nn.L1Loss,
  'logits_sparse_categorical_crossentropy': th.nn.CrossEntropyLoss,
  'smooth_mae'          : th.nn.SmoothL1Loss, # this is similary to HuberLoss
  # TODO: Must add HuberLoss
  # TODO: Must add quantile loss
  }

_OPTIMS  = {
  'adam': th.optim.Adam,
  'sgd': th.optim.SGD,
  'rmsprop':th.optim.RMSprop,
  }

  
  
class ModelTrainer(DecentrAIObject):
  def __init__(self,
               model,
               losses,
               optimizer='adam',
               lr=0.001,
               reload_init_callback=None,
               predict_callback=None,
               evaluate_callback=None,
               evaluate_train_callback=None,
               train_on_batch_callback=None,
               max_patience=10,
               max_fails=30,
               cooldown=2,
               lr_decay=0.5,
               batch_size=32,
               score_mode='max',
               score_key='dev_acc',
               device=None,
               min_score_thr=None,
               min_score_eps=1e-4,
               validation_data=None,
               model_name=None,
               evaluate_train_best_model=True,
               evaluate_train_at_new_best=False,
               base_folder=None,
               **kwargs,
               ):
    
    """
    TODO: Refactor to match tf trainer
    OBS: This is actually the proto-trainer from which emerged the tf trainer

    Parameters
    ----------
    
    model: th.Module
      the model object that must be already instantiated

    loss: callable or str
      loss module class or function. 

    optimizer : str or th.optim, optional
      proposed optimizer class. The default is th.optim.Adam.

    lr : float, optional
      proposed starting learning rate. The default is 0.001.
      
    reload_init_callback: function, optional
      a function that needs to be called after a checkpoint has been called
      
    predict_callback: function, optional
      custom predict function, default is just a forward pass
    
    evaluate_callback: function, optional
      custom evaluate callback, default uses predict and computes accuracy

    train_on_batch_callback: function, optional
      custom train callback, default uses _train_on_batch

    max_patience : int, optional
      max number of epochs for patience. The default is 10.
      
    max_fails : int, optional
      max number of consecutive fails before stopping. The default is 30.
      
    cooldown : int, optional
      how many epochs to skip patience counting. The default is 2.
      
    lr_decay : float, optional
      scaling factor when reducing lr. The default is 0.5.
      
    batch_size : int, optional
      batch size to be used with th.utils.data. The default is 32.
      
    score_mode : str, optional
      how to compare the score key. The default is 'max'.
      
    score_key : str, optional
      the name of the score key. The default is 'acc'.
      
    device : tf.device, optional
      proposed device where all the training will be done. If (default) is None 
      then 'CUDA' will be used.
      
    validation_data : list or tuple of ndarrays, optional
      validation data. The default is None.

    min_score_thr : float, optional
      minimal score to consider the result as valid. The default is None.
      
    min_score_eps : float, optional
      minimal score delta to consider model is better

    base_folder : str, optional
      Folder in which to save the model weights. Defaults to None (self.log.models_folder)

    Returns
    -------
    None.

    """
    self.model = model
    self.model_name = model_name
    self.reload_init = reload_init_callback if reload_init_callback is not None else self._reload_init
    self.predict = predict_callback if predict_callback is not None else self._predict
    self.evaluate = evaluate_callback if evaluate_callback is not None else self._evaluate
    self.evaluate_train = evaluate_train_callback if evaluate_train_callback is not None else self._evaluate
    self._evaluate_train_best_model = evaluate_train_best_model
    self._evaluate_train_at_new_best = evaluate_train_at_new_best
    self.train_on_batch = train_on_batch_callback if train_on_batch_callback is not None else self._train_on_batch

    self._loss_call = losses
    self.losses = None
    self._optimizer_class = optimizer
    self.optimizer = None
    
    self.score_mode = score_mode
    self.score_key = score_key
    self.batch_size = batch_size
    self.device = th.device("cuda" if th.cuda.is_available() else "cpu") if device is None else device
    self.lr = lr        
    self.max_patience = max_patience
    self.max_fails = max_fails
    self.cooldown = cooldown
    self.lr_decay = lr_decay
    self.validation_data = validation_data
    self.min_score_eps = min_score_eps

    if min_score_thr is None:
      self.min_score_thr =  -np.inf if self.score_mode == 'max' else np.inf
    else:
      self.min_score_thr = min_score_thr
    
    self.lr_decay_iters = 0
    self.no_remove=False
    self.errors= []
    self.training_status = {}
    self._files_for_removal = []
    self._debug_data = None
    self.epochs_data = {}
    self.in_training = False
    self.current_epoch = None
    self.score_eval_func = max if self.score_mode == 'max' else min

    self.average_loss = None
    self.base_folder = base_folder

    self._timers_section = 'THModelTrainer'
    super().__init__(**kwargs)

    if self.evaluate_train is None:
      self.P("Warning! No evaluate train callback set!", color='r')
    return
  
  def startup(self):
    super().startup()
    self.__name__ = 'THT'

    if self.base_folder is None:
      self.base_folder = self.log.get_models_folder()
    return

  def P(self, s, **kwargs):
    super().P(s, prefix=True, **kwargs)
    return
  
  def __repr__(self):
    _s  = "  Model: '{}'+\n".format(self.model_name)
    _s += textwrap.indent(str(self.model), " " * 4) + '\n'
    _s += textwrap.indent("Loss: {}".format(self.losses), " " * 4)
    return _s
  
    
  def _evaluate(self, epoch, verbose=False, key='dev', **kwargs):
    ### TODO: Check all list fctions
    if self.validation_data is None:
      raise ValueError('Default evaluate failed due to no validation data')
    dct_results = OrderedDict()
    x_eval = self.validation_data[0]
    if type(x_eval) != list:
      x_eval = [x_eval]
    np_y_hat = self.predict(x_eval)
    if self.score_key.lower() == 'acc':
      n_outs = np_y_hat.shape[-1]
      if n_outs > 1:
        np_y_pred = np.argmax(np_y_hat, -1)
      else:
        np_y_pred = np_y_hat >= 0.5
      y_true = self.validation_data[1]
      dct_results['{}_acc'.format(key)] = (np_y_pred == y_true).sum() / y_true.shape[0]
    else:
      raise ValueError("No default implementation for score type '{}'".format(self.score_key))
    return dct_results
    
    
  def _predict(self, x_data):
    self.model.eval()
    with th.no_grad():
      th_X = [th.tensor(x, device=self.device) for x in x_data]
      if not self.input_type_list:
        th_X = th_X[0]
      th_preds = self.model(th_X)
      np_preds = [x.cpu().numpy() for x in th_preds]
    self.model.train()
    return np.array(np_preds)

    
  def _reload_init(self, dct_epoch_data):
    self.P("Reload Init: {}".format(dct_epoch_data))
    return
  
  def _train_on_batch(self, model, optimizer, losses, batch, y_index=1, batch_index=False):
    # batch_index: flag that specifies if last value of the batch is the batch index
    if batch_index:
      batch = batch[:-1]

    if len(batch) == 1:
      lst_th_x = [batch[0]]
      lst_th_y = None
    else:
      lst_th_x = batch[:y_index]
      lst_th_y = batch[y_index:]
    #endif

    if not self.input_type_list:
      lst_th_x = lst_th_x[0]

    if type(lst_th_x[0]) is list:
      data_device = lst_th_x[0][0].device
    else:
      data_device = lst_th_x[0].device
    #endif
    model_device = next(model.parameters()).device

    if data_device != model_device:
      if type(lst_th_x[0]) is list:
        lst_th_x = [[_x.to(model_device) for _x in x] for x in lst_th_x]
      else:
        lst_th_x = [x.to(model_device) for x in lst_th_x]
      #endif
    #endif


    th_yh = model(lst_th_x)
    if lst_th_y[0].device != th_yh.device:
      lst_th_y = [y.to(th_yh.device) for y in lst_th_y]


    losses_vals = []
    if type(th_yh) is not list:
      th_yh = [th_yh]

    if lst_th_y is not None:
      ### TODO: CHECK If ok
      for i, loss in enumerate(losses):
        losses_vals.append(losses[loss](th_yh[i], lst_th_y[i]))
    else:
      for i, loss in enumerate(losses):
        losses_vals.append(losses[loss](th_yh[i]))
    #endif
    ### TODO: sanity check
    th_loss = th.stack(losses_vals, dim=0).sum(dim=0)#th.sum(*losses_vals, keepdim=True)

    optimizer.zero_grad()
    th_loss.mean().backward()
    optimizer.step()
    err = th_loss.detach().cpu().numpy()
    return err
    

  def _maybe_define_loss(self):
    if self.losses is not None:
      return
    loss_call = None
    if isinstance(self._loss_call, str):
      s_loss = self._loss_call.lower()
      if s_loss in _LOSSES:
        loss_call = _LOSSES[s_loss]
      else:
        raise ValueError("loss '{}' not defined. Available options are: {}".format(
          s_loss, list(_LOSSES.keys())))
    else:
      loss_call = self._loss_call

    if type(loss_call) is dict:
      self.losses = loss_call
    else:
      self.losses = {'loss': loss_call}
    return
  
  def _maybe_define_optimizer(self):
    if self.optimizer is not None:
      return
    optimizer_class = None
    if isinstance(self._optimizer_class,str):
      s_optim = self._optimizer_class.lower()
      if s_optim in _OPTIMS:
        optimizer_class = _OPTIMS[s_optim]
      else:
        raise ValueError('{} not available. Available optimizers are: {}'.format(
          s_optim, list(_OPTIMS.keys())))
    else:
      optimizer_class = self._optimizer_class
    
    self.optimizer = optimizer_class(self.model.parameters(), lr=self.lr)
    return

  def is_new_best(self, dct_score):
    if self.score_key not in dct_score: ## TODO This is in case of train eval
      return False
    score = dct_score[self.score_key]
    test_score = score - self.min_score_eps if self.score_mode.lower() == 'max' else score + self.min_score_eps
    new_best = self.score_eval_func(test_score, self.training_status['best_score']) != self.training_status['best_score']
    return new_best

  def start_timer(self, tmr_id):
    if self.log.is_main_thread:
      section = None
    else:
      section = self._timers_section

    self.log.start_timer(tmr_id, section=section)
    return

  def end_timer(self, tmr_id, skip_first_timing=False, periodic=False):
    if self.log.is_main_thread:
      section = None
    else:
      section = self._timers_section

    self.log.end_timer(tmr_id, skip_first_timing=skip_first_timing, section=section, periodic=periodic)
    return

  def stop_timer(self, tmr_id, skip_first_timing=False, periodic=False):
    self.end_timer(tmr_id=tmr_id, skip_first_timing=skip_first_timing, periodic=periodic)
    return

  def fit(self, x_train=None, th_dl=None, y_train=None, batch_index=False, epochs=10000, verbose=False):
    """
    The training method

    Parameters
    ----------
    x_train : ndarray # TODO: must support list of tensors
      the actual training data.
      
    y_train : ndarray, optional
      y_hat data. The default is None.
      
    epochs : int, optional
      max number of epochs. The default is 10000.
      
    verbose : bool, optional
      show verbose information. The default is False.

    Raises
    ------
    ValueError
      DESCRIPTION.

    Returns
    -------
    TYPE
      DESCRIPTION.

    """    
    if self.model_name is None:
      self.model_name = self.model.__class__.__name__
         
    self._maybe_define_loss()
    self._maybe_define_optimizer()
      
    if self.score_key is None:
      raise ValueError("Scoring config is incomplete!")

    if x_train is None:
      assert th_dl is not None
    if th_dl is None:
      assert x_train is not None
      
    self.P("Moving {} and data to device {}".format(
      self.model.__class__.__name__, self.device))
    self.model.to(self.device)
    if type(x_train) != list:
      x_train = [x_train]
      self.input_type_list = False
    else:
      self.input_type_list = True

    if type(y_train) != list:
      y_train = [y_train]

    if th_dl is None:
      x_tensors = [th.tensor(x, device=self.device) for x in x_train if x is not None]
      y_tensors = [th.tensor(y, device=self.device) for y in y_train if y is not None]

      tensors = x_tensors + y_tensors
      y_index = len(x_tensors)

      th_ds = th.utils.data.TensorDataset(*tensors)
      th_dl = th.utils.data.DataLoader(
          th_ds,
          batch_size=self.batch_size,
          shuffle=True)
    else:
      y_index = -1
    #endif
    self.th_dl = th_dl # TODO: Validate with L
    n_batches = len(th_dl)

    self.P("\nTraining model:\n  {}\n    Training on {} observations with batch size {}.\n".format(
      self,
      n_batches * th_dl.batch_size,
      self.batch_size,
      ))
    patience = 0
    fails = 0
    self.training_status['best_score'] = -np.inf if self.score_mode == 'max' else np.inf
    self.log.P("Training for {} epochs with bs={} and {} steps per epoch - total {} training examples".format(
      epochs, self.batch_size, len(th_dl),  self.batch_size*len(th_dl))
    )
    for epoch in range(1, epochs + 1):
      self.P("Starting epoch {:03d} / {:03d}".format(epoch, epochs))
      epoch_errors = []
      self.in_training = epoch
      self.current_epoch = epoch
      self.start_timer('epoch_time')
      ### TODO: REDO
      enum = enumerate(th_dl)

      mean_time_per_batch = 0
      for i in range(len(th_dl)):
        start_time = time()
        self.start_timer('get_data')
        batch_iter, batch_data = next(enum)
        self.end_timer('get_data')

        self.start_timer('batch_train')
        err = self.train_on_batch(self.model, self.optimizer, self.losses, batch_data, y_index=y_index, batch_index=batch_index)
        epoch_errors.append(err)
        if len(err.shape) > 0:
          self.average_loss = np.mean(np.concatenate(epoch_errors))
        else:
          self.average_loss = np.mean(epoch_errors)
        #endif
        _tmp = batch_iter * mean_time_per_batch
        _tmp += time()-start_time
        _tmp /= (batch_iter + 1)
        mean_time_per_batch = _tmp

        str_log = "Training epoch {:03d} - {:.1f}% (e:{:.1f}s r:{:.1f}s) - avg loss: {:.3f},  Patience {:02d}/{:02d},  Fails {:02d}/{:02d}".format(
          epoch,
          (batch_iter + 1) / n_batches * 100,
          mean_time_per_batch * (batch_iter+1),
          mean_time_per_batch * (n_batches - batch_iter - 1),
          self.average_loss,
          patience, self.max_patience,
          fails, self.max_fails,
        )

        if i == len(th_dl) - 1:
          self.P(str_log)
        else:
          if self.log.is_main_thread:
            Pr(str_log)
        #endif

        self.end_timer('batch_train')
      # end epoch
      self.end_timer('epoch_time', skip_first_timing=False)
      # self.log.show_timers()

      dct_score = self.evaluate(owner=self, epoch=epoch, verbose=verbose)
      # dct_score = {**dct_score_train, **dct_score_eval}
      dct_score['ep'] = epoch
      self.epochs_data[epoch] = dct_score
      score = dct_score[self.score_key]

      if self.is_new_best(dct_score):
        self.P("Found new best score {:.5f} better than {:.5f} at epoch {}.".format(
            score, self.training_status['best_score'], epoch), color='g')

        if self._evaluate_train_at_new_best:
          dct_train_score = self.evaluate_train(owner=self, epoch=epoch, verbose=verbose)
          self.epochs_data[epoch] = {**self.epochs_data[epoch], **dct_train_score}

        self.save_best_model_and_track(epoch, score)
        fails = 0
        patience = 0
      else:
        patience += 1
        if patience > 0:
          fails += 1
        self.P("Finished epoch {:03d}, loss: {:.4f}. Current score {:.5f} < {:.5f}. Patience {:02d}/{:02d},  Fails {:02d}/{:02d}".format(
            epoch, self.average_loss, score, self.training_status['best_score'], patience, self.max_patience, fails, self.max_fails))
        self.P('')
        if patience >= self.max_patience:
          self.P("Patience reached {}/{} - reloading from epoch {} and reducting lr".format(
              patience, self.max_patience, self.training_status['best_epoch']), color='y')
          self.reload_best()
          self.reduce_lr()
          patience = -self.cooldown
        if fails >= self.max_fails:
          self.P("\nMax fails {}/{} reached!".format(fails, self.max_fails))
          break
        #endif
      #endif
    #endfor
    self.restore_best_and_cleanup()

    if self._evaluate_train_best_model:
      dct_train_score = self.evaluate_train(owner=self, epoch=self.training_status['best_epoch'], verbose=verbose)
      self.epochs_data[self.training_status['best_epoch']] = {
        **self.epochs_data[self.training_status['best_epoch']],
        **dct_train_score
      }
    #endif

    self.in_training = None
    return self

  def save_best_model_and_track(self, epoch, score, cleanup=True, verbose=True):
    if not os.path.isdir(self.base_folder):
      os.mkdir(self.base_folder)

    weights_fn = "{}_weights_e{:03}_F{:.2f}.th".format(
            self.model_name, epoch, score
    )

    best_fn = os.path.join(self.base_folder, weights_fn)
    self.save_model(best_fn)
    if verbose:
      self.P("  Saved: '{}'".format(best_fn))
      self.P("  Saved: '{}'".format(best_fn + '.optim'))
    last_best =  self.training_status.get('best_file')
    if last_best != None:
      self._files_for_removal.append(last_best)
      self._files_for_removal.append(last_best + '.optim')
    self.training_status['best_file'] = best_fn
    self.training_status['best_epoch'] = epoch
    self.training_status['best_score'] = score
    if cleanup:
      self.cleanup_files()
    return    
  
  
  def save_model(self, fn):
    th.save(self.model.state_dict(), fn)
    th.save(self.optimizer.state_dict(), fn + '.optim')
    return
   
  def load_model(self, fn):
    self.model.load_state_dict(th.load(fn))
    self.optimizer.load_state_dict(th.load(fn + '.optim'))
    return
    
  
  def reload_best(self, verbose=True):
    self.load_model(self.training_status['best_file'])
    self.reload_init(self.epochs_data[self.training_status['best_epoch']])
    if verbose:
      self.P("  Reloaded '{}' data: {}".format(
          self.training_status['best_file'], 
          dict(self.epochs_data[self.training_status['best_epoch']])))
    return
  
  
  def reduce_lr(self, verbose=True):
    self.lr_decay_iters += 1
    factor = self.lr_decay ** self.lr_decay_iters
    for i, param_group in enumerate(self.optimizer.param_groups):
      lr_old = param_group['lr'] 
      param_group['lr'] = param_group['lr'] * factor
      lr_new = param_group['lr']
    if verbose:
      self.P("  Reduced lr from {:.1e} to {:.1e}".format(lr_old, lr_new))
    return
    


  def cleanup_files(self):
    atmps = 0
    while atmps < 10 and len(self._files_for_removal) > 0:   
      removed = []
      for fn in self._files_for_removal:
        if os.path.isfile(fn):
          try:
            os.remove(fn)
            removed.append(fn)
            self.P("  Removed '{}'".format(fn))
          except:
            pass
        else:
          removed.append(fn)            
      self._files_for_removal = [x for x in self._files_for_removal if x not in removed]
    return
  
  
  def restore_best_and_cleanup(self):
    self.reload_best()
    if self.score_eval_func(self.min_score_thr, self.training_status['best_score']) != self.training_status['best_score']:
      self.P("Best score {:.4f} did not pass minimal threshold of {} - model file will be deleted".format(
          self.training_status['best_score'], self.min_score_thr))
      self._files_for_removal.append(self.training_status['best_file'])
    else:
      self.P("Best score {:.4f} passed minimal threshold of {}".format(
          self.training_status['best_score'], self.min_score_thr))
    self._files_for_removal.append(self.training_status['best_file'] + '.optim')
    self.cleanup_files()
    return
  

if __name__ == '__main__':
  #############################################################################
  #############################################################################
  #############################################################################
  ############################## EXAMPLE ######################################
  #############################################################################
  #############################################################################
  #############################################################################

  import torchvision as tv
  
  from naeural_core import Logger
  from naeural_core.local_libraries.nn.th.layers import InputPlaceholder
  from naeural_core.local_libraries.nn.utils import conv_output_shape

  class SimpleModel(th.nn.Module):
    def __init__(self, 
                 input_shape=(1, 28, 28), 
                 convs=[(8, 3, 2), (32, 3, 2), (64, 3, 2)], 
                 dense=32,
                 **kwargs):
      super().__init__()
      assert len(input_shape) == 3
      self.layers = []
      self.add_layer('input', InputPlaceholder(input_shape))
      input_size = input_shape[0]
      prev_h, prev_w = input_shape[1:]
      for i, (f, k, s) in enumerate(convs):
        cnv = th.nn.Conv2d(
          in_channels=input_size, 
          out_channels=f, 
          kernel_size=k,
          stride=s
          )
        input_size = f
        new_h, new_w = conv_output_shape(
          h_w=(prev_h, prev_w),
          kernel_size=k,
          stride=s,
          )
        self.add_layer('conv_{}'.format(i+1), cnv)
        self.add_layer('reul_{}'.format(i+1), th.nn.ReLU6())
        prev_h = new_h
        prev_w = new_w

      self.add_layer('flatten', th.nn.Flatten())
      input_size = new_h * new_w * f

      # self.add_layer('pre_read', th.nn.Linear(input_size, dense))
      # self.add_layer('pre_read_relu', th.nn.ReLU6())
      # input_size = dense

      self.add_layer('drop',th.nn.Dropout(0.5))
      self.add_layer('readout', th.nn.Linear(input_size, 10))
      return
    
    def forward(self, inputs):
      x = inputs
      for layer in self.layers:
        x = layer(x)
      return x
    
    def predict(self, inputs):
      # we dont have softmax in output above so we make a nice "predict" function
      self.eval()
      with th.no_grad():
        out = self(inputs)
        sm_out = th.nn.Softmax(out)
      self.train()
      return sm_out
    

    def add_layer(self, name, module):
      self.add_module(name, module)      
      self.layers.append(getattr(self, name))
      return
  

  l = Logger('THTST', base_folder='.', app_folder='_local_cache', TF_KERAS=False) 
  
  data_path = l.get_data_folder()
  
  mnist_train = tv.datasets.MNIST(root=data_path, download=True, train=True)
  mnist_test = tv.datasets.MNIST(root=data_path, download=True, train=False)
  
  x_train = mnist_train.data.numpy()
  y_train = mnist_train.targets.numpy()
  
  x_dev = mnist_test.data.numpy()
  y_dev = mnist_test.targets.numpy()

  x_train = (x_train.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  x_dev = (x_dev.reshape((-1, 1, 28, 28)) / 255.).astype('float32')
  y_train = y_train.astype('int64')
  y_dev =y_dev.astype('int64')
  
  l.gpu_info(True)
  
  model = SimpleModel()
  ###TODO: REMOVE DCT
  #model_loss = {'cross_entropy': th.nn.CrossEntropyLoss()}
  model_loss = th.nn.CrossEntropyLoss()
  trainer = ModelTrainer(
    log=l, 
    model=model, 
    losses=model_loss,
    validation_data=(x_dev, y_dev),
    batch_size=256
    )
  trainer.fit(x_train, y_train)
  
  


import os
import torch as th

from naeural_core.serving.base.backends.trt import TensorRTModel
from naeural_core.utils.tracing.onnx.utils import create_from_torch
from naeural_core import Logger

if __name__ == '__main__':
  from ultralytics import YOLO
  from xperimental.th_y8.generate import get_test_images
  from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad

  logger = Logger('TRT', base_folder='.', app_folder='_local_cache')

  device = th.device('cuda:0')
  imgs = get_test_images()
  imgs = th_resize_with_pad(
    img=imgs,
    h=640,
    w=640,
    device=device,
    normalize=True,
    half=True,
    return_original=False
  )[0]
  imgs=imgs.half()

  model = YOLO('yolov8n.pt')
  model = model.model
  path = 'tmp/yolo.engine'
  onnx_path = 'tmp/yolo.onnx'
  try:
    os.unlink(path)
  except Exception:
    pass

  create_from_torch(
    model, device, onnx_path, half=True,
    input_names=['input0'],
    output_names=['output0', 'output1', 'output2', 'output3'],
    args=imgs,
    aggressive_shape_inference=True,
    metadata={
      'precision' : 'fp16'
    }
  )

  trt_model = TensorRTModel(logger)
  trt_model.load_or_rebuild_model(onnx_path, True, 4, device)

  o0, o1, o2, o3 = trt_model(
    imgs[:4],
  )

  trt_model = TensorRTModel(logger)
  trt_model.load_or_rebuild_model(onnx_path, True, 4, device)

  o0, o1, o2, o3 = trt_model(
    imgs[:4],
  )

  print(o0.shape)
  print(o1.shape)
  print(o2.shape)
  print(o3.shape)

  from naeural_core.xperimental.th_y8.utils import Y8, BackendType
  model = YOLO('yolov8n.pt')
  model = model.model
  config = {
    'foo' : 'bar',
    'precision' : 'fp16'
  }
  model = Y8(
    (model, config),
    device,
    topk=False,
    backend_type=BackendType.TENSORRT
  )
  path = 'tmp/yoloy.engine'
  onnx_path = 'tmp/yoloy.onnx'

  try:
    # Make sure we always rebuild this as it is part of the test.
    os.unlink(path)
  except Exception:
    pass

  create_from_torch(
    model, device, onnx_path, half=True,
    input_names=['input0'],
    output_names=['output0', 'output1'],
    args=imgs,
    aggressive_shape_inference=False,
    metadata={
      'precision' : 'fp16'
    }
  )

  # Try this with a reduced batch size
  trt_model = TensorRTModel(logger)
  trt_model.load_or_rebuild_model(onnx_path, True, 8, device)
  o0, o1 = trt_model(imgs[:4])
  for i in range(o0.shape[0]):
    print('==============================')
    for j in range(o1[i].item()):
      print(o0[i, j, :].cpu().numpy())



from time import time
import traceback

from naeural_core.bc import DefaultBlockEngine

class BCWrapper:
  def __init__(self, blockchain_manager : DefaultBlockEngine, owner):
    self.__bc : DefaultBlockEngine = blockchain_manager
    self.__owner = owner
    return
  
  
  def P(self, *args, **kwargs):
    if self.__owner is None:
      print(*args, **kwargs)
    return self.__owner.P(*args, **kwargs)
  
  @property
  def address(self):
    """
    Returns the address of the current node

    Returns
    -------
    str
        The address of the current node in the blockchain
    """
    return self.__bc.address
  
  def sign(
    self, 
    dct_data: dict, 
    add_data: bool = True, 
    use_digest: bool = True, 
    replace_nan: bool = True
  ) -> str:
    """
    Generates the signature for a dict object.
    Does not add the signature to the dict object

    Parameters
    ----------
    dct_data : dict
      the input message as a dict.
      
    add_data: bool, optional
      will add signature and address to the data dict (also digest if required). Default `True`
      
    use_digest: bool, optional  
      will compute data hash and sign only on hash
      
    replace_nan: bool, optional
      will replace `np.nan` and `np.inf` with `None` before signing. 

    Returns
    -------
      text signature

        
      IMPORTANT: 
        It is quite probable that the same sign(sk, hash) will generate different signatures
    """
    return self.__bc.sign(
      dct_data=dct_data,
      add_data=add_data,
      use_digest=use_digest,
      replace_nan=replace_nan
    )
  
  def verify(
    self, 
    dct_data: str, 
    str_signature: str = None, 
    sender_address: str = None,
    return_full_info : bool = True,
  ) -> bool:
    """
    Verifies a signature using the public key of the signer

    Parameters
    ----------
    dct_data : dict
        the data that was signed
        
    str_signature : str
        the base64 encoded signature. Default `None` will be taken from the data
        
    str_signer : str
        the signer's address (string) used as the public key for verification.
        Default `None` will be taken from the data
        
    return_full_info: bool, optional
        whether to return the full verification info. Default `True`      
    

    Returns
    -------
    bool
        True if the signature is valid, False otherwise
    """
    return self.__bc.verify(
      dct_data=dct_data, 
      signature=str_signature, 
      sender_address=sender_address,
      return_full_info=return_full_info,
    )

  def encrypt_str(self, str_data : str, str_recipient : str):
    """
    Encrypts a string using the public key of the recipient using asymmetric encryption

    Parameters
    ----------
    str_data : str
        the data to be encrypted (string)
        
    str_recipient : str
        the recipient's address (string) used as the public key
        
    OBSOLETE:
      compress: bool, optional
          whether to compress the data before encryption. Default `True`

    Returns
    -------
    str
       the base64 encoded encrypted data
    """
    encrypted_data = self.__bc.encrypt(
      plaintext=str_data, receiver_address=str_recipient,
    )
    return encrypted_data
  
  def decrypt_str(self, str_b64data : str, str_sender : str):
    """
    Decrypts a base64 encoded string using the private key of the sender using asymmetric encryption.
    This method is able to decrypt data that was was encrypted for multiple recipients as well.

    Parameters
    ----------
    str_b64data : str
        The base64 encoded encrypted data
        
    str_sender : str
        The sender's address (string) used as the public key for decryption
        
    OBSOLETE:
      embed_compressed: bool, optional
          whether the compression flag is embedded in the data. Default `True`. Modify this only for special cases.

    Returns
    -------
    str
       the decrypted data (string) that can be then decoded to the original data
    """
    decompressed_data = self.__bc.decrypt(
      encrypted_data_b64=str_b64data, sender_address=str_sender,
    )
    return decompressed_data
  
  
  def encrypt_multi(self, str_data : str, lst_recipients : list):
    """
    Encrypts a string using the public key of the recipient using asymmetric encryption

    Parameters
    ----------
    str_data : str
        the data to be encrypted (string)
        
    lst_recipients : list
        the list of recipient's addresses (string) used as the public keys

    Returns
    -------
    str
       the base64 encoded encrypted data
    """
    encrypted_data = self.__bc.encrypt_for_multi(
      plaintext=str_data, receiver_addresses=lst_recipients,
    )
    return encrypted_data


  def get_whitelist(self, with_prefix: bool = False):
    """
    Returns the list of nodes that are allowed to connect to the current node

    Returns
    -------
    list
        The list of addresses that are whitelisted
    """
    lst_allowed = self.__bc.whitelist
    if with_prefix:
      lst_allowed = [self.maybe_add_prefix(addr) for addr in lst_allowed]
    return lst_allowed
  
  
  def get_whitelist_with_names(self):
    """
    Returns the list of nodes that are allowed to connect to the current node with their names

    """
    lst_allowed = self.__bc.whitelist_with_names
    return lst_allowed


  def get_allowed_nodes(self, with_prefix: bool = False):
    """
    Returns the list of nodes that are allowed to connect to the current node. Alias of `get_whitelist`

    Returns
    -------
    list
        The list of addresses that are allowed to connect
    """
    return self.get_whitelist(with_prefix=with_prefix)
  
  
  def is_node_allowed(self, node: str):
    """
    Checks if a node is allowed to connect to the current node

    Parameters
    ----------
    node : str
        The address of the node to check

    Returns
    -------
    bool
        True if the node is allowed, False otherwise
    """
    if node in [None, ""]:
      return False
    return self.__bc.is_allowed(sender_address=node)


  def maybe_remove_addr_prefix(self, address: str):
    """
    Removes the address prefix from the current node's address
    
    Parameters
    ----------
    
    address: str
        The address to remove the prefix from

    Returns
    -------
    str
        The address of the current node without the prefix
    """
    return self.__bc.maybe_remove_prefix(address)
  
  
  def maybe_remove_prefix(self, address: str):
    """
    Removes the prefix from the address. Alias of `maybe_remove_addr_prefix`
    
    Parameters
    ----------
    
    address: str
        The address to remove the prefix from

    Returns
    -------
    str
        The address without the prefix
    """
    return self.maybe_remove_addr_prefix(address)
  
  
  def maybe_add_prefix(self, address: str):
    """
    Adds the prefix to the address
    
    Parameters
    ----------
    
    address: str
        The address to add the prefix to

    Returns
    -------
    str
        The address with the prefix
    """
    return self.__bc._add_prefix(address)
  


### EVM


  @property
  def eth_address(self):
    """
    Returns the EVM address of the current node. 

    Returns
    -------
    str
        The address of the current node in the blockchain
    """
    return self.__bc.eth_address
  

  def eth_sign_message(self, types: list, values: list):
    """
    Signs a message using the EVM account of the current node

    Parameters
    ----------
    types : list
        The types of the values to be signed
        
    values : list
        The values to be signed

    Returns
    -------
    dict
        A dictionary with the following keys
        - types: list
            The types of the values that were signed
        - values: list
            The values that were signed
        - signature: str
            The signature of the data
        - address: str
            The address of the EVM account used to sign the data
    """
    return self.__bc.eth_sign_message(types=types, values=values)  

  def eth_verify_message_signature(
    self, 
    types: list, 
    values: list, 
    signature: str
  ):
    """
    Verifies a message signature using the EVM account of the current node
    
    Parameters
    ----------
    types : list
        The types of the values to be verified
    values : list
        The values to be verified
    signature : str
        The signature to be verified
        
    Returns
    -------
    """
    return self.__bc.eth_verify_message_signature(
      types=types, values=values, signature=signature
    )
  
  def eth_sign_node_epochs(self, node: str, epochs: list, epochs_vals: list, signature_only: bool = True):
    """
    Signs the epochs availabilities for a given node using the EVM account of the current node

    Parameters
    ----------
    node : str
        The address of the node
        
    epochs : list
        The epochs to be signed
        
    epochs_vals : list
        The values of the epochs to be signed
        
    signature_only : bool, optional
        Whether to return only the signature. Default `False`

    Returns
    -------
    dict or just str
        A dictionary with the following keys
        - node: str
            The address of the node
        - epochs_vals: list
            The values of the epochs that were signed
        - eth_signature: str
            The EVM signature of the data
        - eth_address: str
            The address of the EVM account used to sign the data
    """
    return self.__bc.eth_sign_node_epochs(
      node=node, epochs=epochs, epochs_vals=epochs_vals, signature_only=signature_only
    )
    
    
  def node_address_to_eth_address(self, node_address: str):
    """
    Converts a node address to an EVM address

    Parameters
    ----------
    node_address : str
        The address of the node

    Returns
    -------
    str
        The EVM address of the node
    """
    return self.__bc.node_address_to_eth_address(node_address)


  def eth_addr_to_internal_addr(self, eth_node_address):
    return self.__owner.netmon.epoch_manager.eth_to_internal(eth_node_address)
  

  def node_addr_to_eth_addr(self, internal_node_address):
    return self.__bc.node_address_to_eth_address(internal_node_address)


  def eth_addr_list_to_internal_addr_list(self, lst_eth):
    return [self.eth_addr_to_internal_addr(eth) for eth in lst_eth]
  
  
  def get_eth_oracles(self):
    """
    Returns the list of EVM addresses for the known oracles.
    OBS: it does not check if the node is alive or not or if it has a known node address.
    """
    return self.__bc.web3_get_oracles()


  def get_oracles(self, include_eth_addrs: bool = False, wait_interval: int = 15):
    """
    Get the oracles node addresses via the current network smart contract and the
    current available network nodes.
    
    Returns
    -------    
    list, list : oracles, oracles_names
    
    """
    wl, names, eth_oracles = [], [], []
    try:
      eth_oracles = self.get_eth_oracles()
      n_oracles = len(eth_oracles)
      if n_oracles == 0:
        msg = "No oracles found in the smart-contracts!"
        raise Exception(msg)
      found = []
      min_converted_thr = int(n_oracles * 0.5)
      _done = False
      _check_start = time()
      while not _done:
        for eth_addr in eth_oracles:
          if eth_addr in found:
            continue
          internal_addr = self.eth_addr_to_internal_addr(eth_addr)
          if internal_addr is None:
            continue
          found.append(eth_addr)
          wl.append(internal_addr)
          alias = self.__owner.netmon.network_node_eeid(internal_addr)
          names.append(alias)
        #end for
        if len(found) >= min_converted_thr or time() - _check_start > wait_interval:
          _done = True
        else:
          self.P("Waiting for oracles to be converted...", color='y')
          time.sleep(2)
        #end if
      #end while
      if len(found) < min_converted_thr:
        msg = "Not enough oracles found to internal addresses. Retrieved from smart contract: {} / Found: {}".format(
          eth_oracles, found
        )
        raise Exception(msg)
    except Exception as e:
      self.P(f"Error getting whitelist data: {e}\n{traceback.format_exc()}", color='r')
    if include_eth_addrs:
      return wl, names, found
    return wl, names  
  
  def is_node_licensed(self, node_address: str = None, node_address_eth: str = None):
    """
    Checks if a node is licensed

    Parameters
    ----------
    node_address : str, optional
        The address of the node, if not provided, the current node's address will be used
        
    node_address_eth : str, optional
        The EVM address of the node to be used instead of node_address

    Returns
    -------
    bool
        True if the node is licensed, False otherwise
    """
    if node_address is None and node_address_eth is None:
      # using the current node's eth address
      node_address_eth = self.__bc.eth_address
    elif node_address_eth is None and node_address is not None:
      node_address_eth = self.node_address_to_eth_address(node_address)
    is_licensed = self.__bc.web3_is_node_licensed(address=node_address_eth)
    return is_licensed
  
  def get_evm_network(self):
    """
    Get the EVM network

    Returns
    -------
    str
        The EVM network
    """
    return self.__bc.evm_network
  
  
  def get_node_license_info(
    self, 
    node_address: str = None, 
    node_address_eth: str = None,
    raise_if_error: bool = False,
  ):
    """
    Get the license info for a node

    Parameters
    ----------
    node_address : str, optional
        The address of the node, if not provided, the current node's address will be used
        
    node_address_eth : str, optional
        The EVM address of the node to be used instead of node_address

    Returns
    -------
    dict
        A dictionary with the license info for the node
    """
    if node_address is None and node_address_eth is None:
      # using the current node's eth address
      node_address_eth = self.__bc.eth_address
    elif node_address_eth is None and node_address is not None:
      node_address_eth = self.node_address_to_eth_address(node_address)
    return self.__bc.web3_get_node_info(
      node_address=node_address_eth, raise_if_issue=raise_if_error      
    )
    
    
  def eth_sign_payload(self, payload: dict, add_data: bool = True):
    """
    Signs a payload using the EVM account of the current node

    Parameters
    ----------
    payload : dict
        The payload to be signed
        
    add_data : bool, optional
        Whether to add the signature and address to the payload. 
        Default `True` will add the signature and address to the payload

    Returns
    -------
    str
        The signature of the payload
    """
    return self.__bc.eth_sign_payload(payload)
  
  
  def eth_verify_payload_signature(self, payload: dict, message_prefix: str = "", no_hash: bool = False, indent=0):
    """
    Verifies a payload signature using the EVM account of the current node

    Parameters
    ----------
    
    payload : dict
        The payload that was signed. Must contain the keys ETH_SENDER and ETH_SIGN.
        
    no_hash : bool, optional
        If True, the message is not hashed before verification. The default is False.
        This is useful for raw text messages that are not hashed such as the ones signed by
        wallets.
        
    message_prefix : str, optional
        A prefix to be added to the message before hashing (or signing). The default is "".
        
    
    indent : int, optional
        The indentation level for the JSON string. The default is 0.
        

    Returns
    -------
    str
        The address of the signer or None if the signature is invalid
    """
    return self.__bc.eth_verify_payload_signature(
      payload=payload, message_prefix=message_prefix, 
      no_hash=no_hash, indent=indent
    )
  
  @property
  def eth_types(self):
    """
    Class with supported EVM types

    Returns
    -------
    list
        The types of the EVM account
    """
    return self.__bc.eth_types
  
  
  def is_valid_eth_address(self, address: str):
    """
    Checks if the address is a valid EVM address

    Parameters
    ----------
    address : str
        The address to be checked

    Returns
    -------
    bool
        True if the address is a valid EVM address, False otherwise
    """
    return self.__bc.is_valid_eth_address(address)
  
  
  def address_is_valid(self, address: str):
    """
    Checks if the address is a valid internal ddress

    Parameters
    ----------
    address : str
        The address to be checked

    Returns
    -------
    bool
        True if the address is a valid address, False otherwise
    """
    return self.__bc.address_is_valid(address)
  
  
  def is_valid_internal_address(self, address: str):
    """
    Checks if the address is a valid internal address

    Parameters
    ----------
    address : str
        The address to be checked

    Returns
    -------
    bool
        True if the address is a valid internal address, False otherwise
    """
    return self.address_is_valid(address)
  
  
  def get_wallet_nodes(self, address: str):
    """
    Get the nodes based on the licenses owned by a given address

    Parameters
    ----------
    address : str
        The address of the wallet

    Returns
    -------
    list
        A list of nodes that are associated with the wallet address
    """
    return self.__bc.web3_get_wallet_nodes(address=address)


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 16:29:02 2020

@author: Andrei
"""


from .utils import get_random_series, moving_averages, get_series_trends, get_trend_preds, get_linear_preds
from .utils import acf, acf_limit, acf_plot, get_valid_lags, get_max_lag
from .utils import plot_autoreg
from .utils import get_daily_series_confidence
from .baselines import get_avail_baselines, get_preconfigured_baselines
from .benchmarker import TimeseriesBenchmarker